/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/
package de.schmitzm.jfree;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Stroke;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.LegendItemSource;
import org.jfree.chart.annotations.XYTitleAnnotation;
import org.jfree.chart.axis.Axis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.block.BlockBorder;
import org.jfree.chart.block.BlockContainer;
import org.jfree.chart.block.BorderArrangement;
import org.jfree.chart.block.ColumnArrangement;
import org.jfree.chart.block.EmptyBlock;
import org.jfree.chart.event.PlotChangeEvent;
import org.jfree.chart.event.PlotChangeListener;
import org.jfree.chart.event.RendererChangeEvent;
import org.jfree.chart.event.RendererChangeListener;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.labels.XYToolTipGenerator;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PiePlot;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.AbstractRenderer;
import org.jfree.chart.renderer.category.AbstractCategoryItemRenderer;
import org.jfree.chart.renderer.category.AreaRenderer;
import org.jfree.chart.renderer.category.BarRenderer;
import org.jfree.chart.renderer.category.CategoryItemRenderer;
import org.jfree.chart.renderer.category.CategoryStepRenderer;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.chart.renderer.category.StackedAreaRenderer;
import org.jfree.chart.renderer.category.StackedBarRenderer;
import org.jfree.chart.renderer.xy.AbstractXYItemRenderer;
import org.jfree.chart.renderer.xy.StackedXYAreaRenderer2;
import org.jfree.chart.renderer.xy.StackedXYBarRenderer;
import org.jfree.chart.renderer.xy.XYAreaRenderer;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.renderer.xy.XYStepAreaRenderer;
import org.jfree.chart.renderer.xy.XYStepRenderer;
import org.jfree.chart.title.CompositeTitle;
import org.jfree.chart.title.LegendTitle;
import org.jfree.chart.title.TextTitle;
import org.jfree.chart.title.Title;
import org.jfree.chart.urls.StandardXYURLGenerator;
import org.jfree.chart.urls.XYURLGenerator;
import org.jfree.data.Range;
import org.jfree.data.UnknownKeyException;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.general.PieDataset;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.ui.HorizontalAlignment;
import org.jfree.ui.RectangleAnchor;
import org.jfree.ui.RectangleEdge;
import org.jfree.ui.RectangleInsets;

import de.schmitzm.jfree.chart.FilteredLegendItemSource;
import de.schmitzm.jfree.chart.renderer.SelectionRenderer;
import de.schmitzm.jfree.chart.selection.DatasetSelectionModel;
import de.schmitzm.jfree.chart.style.BasicChartStyle;
import de.schmitzm.jfree.chart.style.ChartStyle;
import de.schmitzm.jfree.chart.style.ChartStyleXMLFactory;
import de.schmitzm.jfree.chart.style.ChartType;
import de.schmitzm.jfree.data.RegressionDataset;
import de.schmitzm.jfree.data.RegressionDatasetImpl;
import de.schmitzm.lang.LangUtil;
import de.schmitzm.lang.ResourceProvider;

/**
 * In dieser Klasse sind Hilfsmethoden fuer JFreeChart hinterlegt.
 * 
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class JFreeChartUtil {
	private static Logger LOGGER = Logger.getLogger(JFreeChartUtil.class
			.getName());

	/**
	 * Default percentage, an (axis) range is extended with on centering to
	 * avoid points directly on the plot border.
	 */
	public static final Double DEFAULT_CENTRATION_BOUNDARY_PCT = 0.1;

	/**
	 * {@link ResourceProvider}, der die Lokalisation fuer GUI-Komponenten zur
	 * Verfuegung stellt. Diese sind in properties-Datein unter {@code
	 * schmitzm.jfree.resource.locales.JFreeResourceBundle_XXX.properties}
	 * hinterlegt.
	 */
	public static ResourceProvider RESOURCE = ResourceProvider.newInstance(LangUtil
			.extendPackagePath(JFreeChartUtil.class,
					"resource.locales.JFreeResourceBundle"), Locale.ENGLISH);

	/**
	 * Convenience method to access the {@link ResourceProvider}.
	 */
	public static String R(String key, Object... values) {
		return RESOURCE.getString(key, values);
	}

	/** Instance of {@link ChartStyleXMLFactory}. */
	public static final ChartStyleXMLFactory<ChartStyle> CHART_STYLE_FACTORY = new ChartStyleXMLFactory<ChartStyle>();

	/** Prefix for dataset series key which should be filtered out in
	 *  legend.
	 *  @see FeatureChartUtil#createCategoryDataset(org.geotools.feature.FeatureCollection, schmitzm.jfree.feature.style.FeatureChartStyle)
	 *  @see BasicChartStyle#applyToDataset(Dataset...)
	 *  */
	public static final String DUMMY_SERIES_PREFIX = "DUMMY_"; 
	

	/**
	 * Sets all chart legends (in)visible.
	 * 
	 * @param chart
	 *            a chart to apply
	 * @param visible
	 *            indicates the visible property
	 */
	public static void setLegendVisible(JFreeChart chart, boolean visible) {
//		for (Object title : chart.getSubtitles())
//			if (title instanceof LegendTitle)
//				((Title) title).setVisible(visible);
        for (Object title : chart.getSubtitles()) {
          if (title instanceof LegendTitle)
            ((Title) title).setVisible(visible);
          // If legend is split because of multiple range axis, there
          // is an CompositeTitle instead of LegendTitle. Setting
          // (in)visible the LegendTitles inside, does not take effect.
          // -> So set the whole CompositeTitle (in)visible if one
          //    of the blocks inside is a LegendTitle...
          if (title instanceof CompositeTitle) {
            for (Object block : ((CompositeTitle)title).getContainer().getBlocks()) {
              if ( block instanceof LegendTitle ) {
                ((CompositeTitle)title).setVisible(visible);
                break;
              }
            }
          }
        }
	}
	
	/**
	 * Splits the legend of a chart in two pieces for each range
	 * axis.<br>
	 * Does nothing if chart has only 1 range axis.
	 */
  	public static void splitLegend(JFreeChart chart) {
  	  if ( getRangeAxisCountFromPlot(chart.getPlot()) < 2 )
  	    return;
  	  
  	  // remove old legend
  	  chart.removeLegend();

      // create left legend for first dataset legend items 
  	  LegendItemSource lis1 = new FilteredLegendItemSource(
  	      (LegendItemSource)getRendererForDataset(chart.getPlot(), 0)
  	  );
  	  LegendTitle legend1 = new LegendTitle(lis1);
  	  legend1.setMargin(new RectangleInsets(2, 2, 2, 2));
      // Create border, only if legend items are present (avoid
      // border around empty legend, because it is shown as
      // vertical line)
      if ( lis1.getLegendItems().getItemCount() > 0 )
        legend1.setFrame(new BlockBorder());
  	
      // create right legend for second dataset legend items 
  	  LegendItemSource lis2 = new FilteredLegendItemSource(
  	      (LegendItemSource)JFreeChartUtil.getRendererForDataset(chart.getPlot(), 1)
  	  );   
  	  LegendTitle legend2 = new LegendTitle(lis2);
  	  legend2.setMargin(new RectangleInsets(2, 2, 2, 2));
  	  // Create border, only if legend items are present (avoid
  	  // border around empty legend, because it is shown as
  	  // vertical line)
  	  if ( lis2.getLegendItems().getItemCount() > 0 )
  	    legend2.setFrame(new BlockBorder());
    	  
  	  // combine the legends in a block
  	  BlockContainer container = new BlockContainer(new BorderArrangement());
      container.add(legend1, RectangleEdge.LEFT);
  	  container.add(legend2, RectangleEdge.RIGHT);
  	  container.add(new EmptyBlock(2000, 0));
  	  
  	  // create the new legend
  	  CompositeTitle legends = new CompositeTitle(container);
  	  legends.setPosition(RectangleEdge.BOTTOM);
  	  chart.addSubtitle(legends);
  	}
  	
    /**
     * When all datasets for a range axis are empty, JFreeChart
     * shows a default range with tick labels between 0.0 and 1.0.
     * That looks not nice for us (especially Stefan!). We (Stefan) want
     * no tick labels in this case.<br>
     * This method first disables all tick labels for all axis. Then
     * it checks all datasets and (if not empty) (re)enable the tick labels
     * for the corresponding range axis
     * @param plot the plot the tick labels should be reset for
     */
    public static void resetRangeAxisTickLabelVisibiliy(JFreeChart chart) {
      if ( chart != null )
        resetRangeAxisTickLabelVisibiliy(chart.getPlot());
    }
    
    /**
  	 * When all datasets for a range axis are empty, JFreeChart
  	 * shows a default range with tick labels between 0.0 and 1.0.
  	 * That looks not nice for us (especially Stefan!). We (Stefan) want
  	 * no tick labels in this case.<br>
  	 * This method first disables all tick labels for all axis. Then
  	 * it checks all datasets and (if not empty) (re)enable the tick labels
  	 * for the corresponding range axis
  	 * @param plot the plot the tick labels should be reset for
  	 */
  	public static void resetRangeAxisTickLabelVisibiliy(Plot plot) {
  	  if ( plot == null )
  	    return;
  	  
      int datasetCount = JFreeChartUtil.getDatasetCountFromPlot(plot);
      // first disable the tick labels for all range axis
      for (int i=0; i<datasetCount; i++) {
        Dataset dataset = JFreeChartUtil.getDatasetFromPlot(plot, i);
        ValueAxis axis = JFreeChartUtil.getRangeAxisForDataset(plot, dataset);
        if ( axis != null )
          axis.setTickLabelsVisible(false);
      }
      // (re)enable the tick labels for the axis whose dataset(s) are
      // not empty
      for (int i=0; i<datasetCount; i++) {
        Dataset dataset = JFreeChartUtil.getDatasetFromPlot(plot, i);
        ValueAxis axis = JFreeChartUtil.getRangeAxisForDataset(plot, dataset);
        if ( axis != null && !JFreeChartUtil.isDatasetEmpty(dataset, true) )
          axis.setTickLabelsVisible(true);
      }
  	}
  	  
	/**
	 * Fires a change event to all {@link PlotChangeListener} connected to the
	 * plot.
	 * 
	 * @param plot
	 *            a plot
	 */
	public static void fireChangeEvent(Plot plot) {
		plot.notifyListeners(new PlotChangeEvent(plot));
	}

	/**
	 * Fires a change event to all {@link RendererChangeListener} connected to
	 * the renderer.
	 * 
	 * @param renderer
	 *            a renderer
	 */
	public static void fireChangeEvent(AbstractRenderer renderer) {
		renderer.notifyListeners(new RendererChangeEvent(renderer, true));
	}

	/**
	 * Returns the maximum extend of a {@link Range} including a given boundary.
	 * 
	 * @param range
	 *            a {@link Range}
	 * @param boundaryPct
	 *            extends the {@link Range} bound by percentage (if {@code null}
	 *            {@value #DEFAULT_CENTRATION_BOUNDARY_PCT} is used).
	 * @return a value >= 0
	 */
	public static double getAbsRange(Range range, Double boundaryPct) {
		if (boundaryPct == null)
			boundaryPct = DEFAULT_CENTRATION_BOUNDARY_PCT;
		return Math.max(Math.abs(range.getUpperBound()), Math.abs(range
				.getLowerBound()))
				* (1 + boundaryPct);
	}

	/**
	 * Returns a new centered {@link Range} defined as [-maxAbs; maxAbs] of a
	 * given {@link Range}.
	 * 
	 * @param range
	 *            the {@link Range} to center
	 * @param boundaryPct
	 *            extends the range's absolute maximum by percentage (if {@code
	 *            null} {@value #DEFAULT_CENTRATION_BOUNDARY_PCT} is used).
	 * @return a new {@link Range} object
	 */
	public static Range centerRange(Range range, Double boundaryPct) {
		double absMax = getAbsRange(range, boundaryPct);
		return new Range(-absMax, absMax);
	}

	/**
	 * Centers a gives {@link Axis} around 0 (if normalized) or around it's
	 * center.
	 * 
	 * @param axis
	 *            the {@link Axis} to center
	 * @param boundaryPct
	 *            extends the axis range's absolute maximum by percentage (if
	 *            {@code null} {@value #DEFAULT_CENTRATION_BOUNDARY_PCT} is
	 *            used).
	 * @param symmetric
	 *            If <code>true</code> the axis is centered symetrically
	 */
	public static void zoomAxisToFullExtend(ValueAxis axis, Double boundaryPct,
			boolean symmetric) {
		// first recalculate the full extend (unfortunately the listeners are
		// called; the method to de-activate this behavior is protected!)
		axis.setAutoRange(true);
		// then reset the centering
		if (symmetric) {
			double absMax = getAbsRange(axis.getRange(), boundaryPct);
			axis.setRange(-absMax, absMax);
		} else {
			axis.setRange(axis.getRange().getLowerBound(), axis.getRange()
					.getUpperBound());
		}
	}

	/**
	 * Centers a gives {@link Plot}. For {@link CategoryPlot} only the range
	 * axis is centered.
	 * 
	 * @param axis
	 *            the {@link Plot} to center
	 * @param boundaryPct
	 *            percentage boundary the absolute maximum is extended with
	 * @param symetric
	 *            If <code>true</code> the axis is centered symmetrically around
	 *            it medium.
	 */
	public static void zoomPlotToFullExtend(Plot plot, Double boundaryPct,
			boolean symetric) {
		if (plot instanceof XYPlot) {
			zoomAxisToFullExtend(((XYPlot) plot).getDomainAxis(), boundaryPct,
					symetric);
			zoomAxisToFullExtend(((XYPlot) plot).getRangeAxis(), boundaryPct,
					symetric);
		}
		if (plot instanceof CategoryPlot)
			zoomAxisToFullExtend(((CategoryPlot) plot).getRangeAxis(),
					boundaryPct, symetric);
	}

	/**
	 * Creates a {@link XYItemRenderer} for the given stepped and stacked
	 * properties.
	 * 
	 * @param type
	 *            chart type
	 * @param stepped
	 *            stepped property
	 * @param stacked
	 *            stacked property
	 */
	public static XYItemRenderer createXYRenderer(BasicChartStyle style) {
		return createXYRenderer(style.getType(), style.isStepped(), style
				.isStacked());
	}

	/**
	 * Creates a {@link XYItemRenderer} for the given stepped and stacked
	 * properties.
	 * 
	 * @param type
	 *            chart type
	 * @param stepped
	 *            stepped property
	 * @param stacked
	 *            stacked property
	 */
	public static XYItemRenderer createXYRenderer(ChartType type,
			boolean stepped, boolean stacked) {
		switch (type) {
		case LINE: // "stacked" is not available for line charts
			if (stacked)
				LOGGER
						.warn("Stacked property can not be applied to line charts. Property ignored!");
			// Create renderer according to "stepped"
			if (stepped)
				return new XYStepRenderer();
			// normal line renderer
			return new XYLineAndShapeRenderer(true, false);
		case AREA: // Apply stacked
			if (stacked) {
				// "stacked" and "stepped" can not be applied the same time
				if (stepped)
					LOGGER
							.warn("A stacked area chart can not be stepped the same time (use a stacked bar chart instead). Stepped propery ignored!");
				// Stacked area renderer
				return new StackedXYAreaRenderer2();
			}
			// Stepped area renderer
			if (stepped)
				return new XYStepAreaRenderer(XYStepAreaRenderer.AREA);
			// Neither stepped nor stepped
			return new XYAreaRenderer(XYAreaRenderer.AREA);
		case BAR: // bar charts are automatically stepped,
			// so only "stacked" is applied
			if (stacked)
				return new StackedXYBarRenderer();
			// Normal bar renderer
			return new XYBarRenderer();
		}
		throw new UnsupportedOperationException(
				"XYItemRenderer could not be created for chart type: " + type);
	}

	/**
	 * Creates a {@link CategoryItemRenderer} for the given stepped and stacked
	 * properties.
	 * 
	 * @param type
	 *            chart type
	 * @param stepped
	 *            stepped property
	 * @param stacked
	 *            stacked property
	 */
	public static CategoryItemRenderer createCategoryRenderer(
			BasicChartStyle style) {
		return createCategoryRenderer(style.getType(), style.isStepped(), style
				.isStacked());
	}

	/**
	 * Creates a {@link CategoryItemRenderer} for the given stepped and stacked
	 * properties.
	 * 
	 * @param type
	 *            chart type
	 * @param stepped
	 *            stepped property
	 * @param stacked
	 *            stacked property
	 */
	public static CategoryItemRenderer createCategoryRenderer(ChartType type,
			boolean stepped, boolean stacked) {
		switch (type) {
		case LINE: // "stacked" is not available for line charts
			if (stacked)
				LOGGER
						.warn("Stacked property can not be applied to line charts. Property ignored!");
			// Create renderer according to "stepped"
			if (stepped)
				return new CategoryStepRenderer();
			// normal line renderer
			return new LineAndShapeRenderer(true, false);
		case AREA: // category area chart is automatically stepped,
			// so only "stacked" is applied
			if (stacked)
				return new StackedAreaRenderer(false);
			// Normal area renderer
			return new AreaRenderer();
		case BAR: // bar charts are automatically stepped,
			// so only "stacked" is applied
			if (stacked)
				return new StackedBarRenderer(false);
			// Normal bar renderer
			return new BarRenderer();
		}
		throw new UnsupportedOperationException(
				"CategoryItemRenderer could not be created for chart type: "
						+ type);
	}

    /**
     * Returns the plot from a renderer.
     * @param renderer a renderer
     */
    public static Plot getPlotFromRenderer(AbstractRenderer renderer) {
      if ( renderer instanceof AbstractCategoryItemRenderer )
        return ((AbstractCategoryItemRenderer)renderer).getPlot();
      if ( renderer instanceof AbstractXYItemRenderer )
        return ((AbstractXYItemRenderer)renderer).getPlot();
      LOGGER.warn("Could not determine plot for renderer: "+LangUtil.getSimpleClassName(renderer));
      return null;
    }
    
    
    /**
     * Returns range axis count from a plot.
     * @param plot a plot
     */
    public static int getRangeAxisCountFromPlot(Plot plot) {
      if ( plot instanceof XYPlot )
        return ((XYPlot)plot).getRangeAxisCount();
      if ( plot instanceof CategoryPlot )
        return ((CategoryPlot)plot).getRangeAxisCount();
      if ( plot instanceof PiePlot )
        return 0;
      LOGGER.warn("Could not determine range axis count for plot: "+LangUtil.getSimpleClassName(plot));
      return 0;
    }

    public static int getDatasetIndex(Plot plot, Dataset dataset) {
      int idx = -1;
      if ( plot instanceof CategoryPlot ) {
        CategoryPlot catPlot = (CategoryPlot)plot;
        for (int i=0; i<catPlot.getDatasetCount(); i++)
          if ( dataset == catPlot.getDataset(i) ) {
            idx = i;
            break;
          }
      } else if ( plot instanceof XYPlot ) {
        XYPlot xyPlot = (XYPlot)plot;
        for (int i=0; i<xyPlot.getDatasetCount(); i++)
          if ( dataset == xyPlot.getDataset(i) ) {
            idx = i;
            break;
          }
      } else if ( plot instanceof PiePlot ) {
        PiePlot piePlot = (PiePlot)plot;
        if ( dataset == piePlot.getDataset() )
          idx = 0;
      } else
        LOGGER.warn("Plot does not contain dataset!");
      
      return idx;
    }
    
    /**
     * Returns range axis count from a plot.
     * @param plot a plot
     */
    public static ValueAxis getRangeAxisForDataset(Plot plot, Dataset dataset) {
      int idx = getDatasetIndex(plot, dataset);
      
      if ( plot instanceof XYPlot )
        return ((XYPlot)plot).getRangeAxisForDataset(idx);
      if ( plot instanceof CategoryPlot )
        return ((CategoryPlot)plot).getRangeAxisForDataset(idx);
      
      LOGGER.warn("Could not determine range axis for plot: "+LangUtil.getSimpleClassName(plot));
      return null;
    }

    /**
     * Returns domain axis count from a plot.
     * @param plot a plot
     */
    public static int getDomainAxisCountFromPlot(Plot plot) {
      if ( plot instanceof XYPlot )
        return ((XYPlot)plot).getDomainAxisCount();
      if ( plot instanceof CategoryPlot )
        return ((CategoryPlot)plot).getDomainAxisCount();
      if ( plot instanceof PiePlot )
        return 0;
      LOGGER.warn("Could not determine domain axis count for plot: "+LangUtil.getSimpleClassName(plot));
      return 0;
    }

    /**
     * Returns dataset count from a plot.
     * @param plot a plot
     */
    public static int getDatasetCountFromPlot(Plot plot) {
      if ( plot instanceof XYPlot )
        return ((XYPlot)plot).getDatasetCount();
      if ( plot instanceof CategoryPlot )
        return ((CategoryPlot)plot).getDatasetCount();
      if ( plot instanceof PiePlot )
        return ((PiePlot)plot).getDataset() != null ? 1 : 0;
      LOGGER.warn("Could not determine dataset count for plot: "+LangUtil.getSimpleClassName(plot));
      return 0;
    }

    /**
     * Returns the dataset from a plot.
     * @param plot a plot
     */
    public static Dataset getDatasetFromPlot(Plot plot, int... datasetIdx) {
      int idx = datasetIdx.length > 0 ? datasetIdx[0] : 0;
      
      if ( plot instanceof XYPlot )
        return ((XYPlot)plot).getDataset(idx);
      if ( plot instanceof CategoryPlot )
        return ((CategoryPlot)plot).getDataset(idx);
      if ( plot instanceof PiePlot )
        return ((PiePlot)plot).getDataset();
      LOGGER.warn("Could not determine dataset for plot: "+LangUtil.getSimpleClassName(plot));
      return null;
    }

    /**
     * Returns the dataset from a renderer or plot.
     * @param renderer a renderer (also 
     */
    public static Dataset getDatasetFrom(Object object) {
      Dataset dataset = null;
      if ( object instanceof AbstractRenderer )
        object = getPlotFromRenderer((AbstractRenderer)object);
      if ( object instanceof Plot )
        dataset = getDatasetFromPlot((Plot)object);
      return dataset;
    }
    
    
    /**
     * Checks whether a dataset is empty.
     * @param dataset a dataset
     */
    public static boolean isDatasetEmpty(Dataset dataset, boolean checkEveryItemForNull) {
      if ( dataset instanceof XYDataset ) {
        XYDataset xyDataset = (XYDataset)dataset;
        for (int seriesIdx=0; seriesIdx<xyDataset.getSeriesCount(); seriesIdx++)
          if ( !checkEveryItemForNull ) {
            if ( xyDataset.getItemCount(seriesIdx) > 0 )
              return false;
          } else {
            for (int itemIdx=0; itemIdx<xyDataset.getItemCount(seriesIdx); itemIdx++)
             if ( xyDataset.getY(seriesIdx, itemIdx) != null )
               return false;
          }
        return true;
      } else if ( dataset instanceof CategoryDataset ) {
        CategoryDataset catDataset = (CategoryDataset)dataset;
        if ( !checkEveryItemForNull )
          return catDataset.getRowCount() > 0;
        for (int col=0; col<catDataset.getColumnCount(); col++)
          for (int row=0; row<catDataset.getRowCount(); row++)
            if ( catDataset.getValue(row, col) != null )
              return false;
      } else if ( dataset instanceof PieDataset ) {
        PieDataset pieDataset = (PieDataset)dataset;
        if ( !checkEveryItemForNull )
          return pieDataset.getItemCount() > 0;
        for ( int j=0; j<pieDataset.getItemCount(); j++ )
          if ( pieDataset.getValue(j) != null )
            return false;
        return true;
      } else
        LOGGER.warn("Could not determine emptyness for dataset: "+LangUtil.getSimpleClassName(dataset));

      return true;
    }

    /**
     * Returns the series index for a series key in a dataset.
     * For {@link PieDataset} this method always returns 0 (unless
     * series key is {@code null}), because pies have only one
     * series!
     * @param dataset a dataset
     * @param seriesKey a series key 
     * @return -1 if there is no series with the key
     */
    public static int getSeriesIndexFromDataset(Dataset dataset, Comparable seriesKey) {
      if ( dataset instanceof CategoryDataset )
        return ((CategoryDataset)dataset).getRowIndex(seriesKey);
      if ( dataset instanceof XYDataset ) {
        XYDataset xyDataset = (XYDataset)dataset;
        for (int i=0; i<xyDataset.getSeriesCount(); i++)
          if ( seriesKey.equals( xyDataset.getSeriesKey(i) ) )
            return i;
        // no series available for key
        return -1;
      }
      if ( dataset instanceof PieDataset )
        // PieDataset has only one series, so return this
        // series for every key
        return seriesKey != null ? 0 : -1;
      
      LOGGER.warn("Could not determine series index for dataset: "+LangUtil.getSimpleClassName(dataset));
      return -1;
    }

    /**
     * Returns the series count from a dataset.
     * For {@link PieDataset} this method always returns 1 because pies
     * have only one series!
     * @param dataset a dataset
     * @return 0 if there is no series
     */
    public static int getSeriesCountFromDataset(Dataset dataset) {
      if ( dataset instanceof CategoryDataset )
        return ((CategoryDataset)dataset).getRowCount();
      if ( dataset instanceof XYDataset )
        return ((XYDataset)dataset).getSeriesCount();
      if ( dataset instanceof PieDataset )
        return 1;
      
      LOGGER.warn("Could not determine series count for dataset: "+LangUtil.getSimpleClassName(dataset));
      return 0;
    }

    /**
     * Returns the series key for a series in a dataset.
     * For {@link PieDataset} this method always returns an empty string (unless
     * an invalid series < 0 is given), because pies have only one series!
     * @param dataset a dataset
     * @param series a series key 
     * @return {@code null} if there is no series for the index
     */
    public static Comparable getSeriesKeyFromDataset(Dataset dataset, int series) {
      try {
        if ( dataset instanceof CategoryDataset )
          return ((CategoryDataset)dataset).getRowKey(series);
        if ( dataset instanceof XYDataset )
          return ((XYDataset)dataset).getSeriesKey(series);
        if ( dataset instanceof PieDataset )
          return series >= 0 ? "" : null;
      } catch (IndexOutOfBoundsException err) {
        return null;
      }
      LOGGER.warn("Could not determine series key for dataset: "+LangUtil.getSimpleClassName(dataset));
      return null;
    }

    /**
	 * Returns the renderer for the {@code i}-th dataset of a plot.
	 * 
	 * @param plot
	 *            a plot
	 * @param index
	 *            the dataset index
	 * @return the primary renderer if no specific renderer is set for dataset
	 *         {@code i}
	 */
	public static Object getRendererForDataset(Plot plot, int i) {
		Object renderer = null;
		if (plot instanceof XYPlot) {
			renderer = ((XYPlot) plot).getRenderer(i);
			if (renderer == null)
				renderer = ((XYPlot) plot).getRenderer();
		}

		if (plot instanceof CategoryPlot) {
			renderer = ((CategoryPlot) plot).getRenderer(i);
			if (renderer == null)
				renderer = ((CategoryPlot) plot).getRenderer();
		}

		return renderer;
	}

	/**
	 * Returns a series from a dataset. If no series exists a new one is created.
	 * @param dataset   the dataset
	 * @param seriesKey the key to return the series for
	 * @param autoSort  indicates the "autosort" property for the series (if a new series is created)
	 * @param allowDupl indicates whether duplicate x values are allowed in the series (if a new series is created)
	 * @return never {@code null}
	 */
	public static XYSeries getOrAddSeriesFromDataset(XYSeriesCollection dataset, Comparable seriesKey, boolean autoSort, boolean allowDupl) {
	  try {
	    return dataset.getSeries(seriesKey);
	  } catch (UnknownKeyException err) {
	    XYSeries xySeries = new XYSeries(seriesKey, autoSort, allowDupl); // IMPORTANT: Do not use autosort!!
	    dataset.addSeries( xySeries );
	    return xySeries;
	  }
	}

    /**
     * Returns a series from a dataset. If no series exists a new one is created.
     * @param dataset   the dataset
     * @param seriesKey the key to return the series for
     * @return never {@code null}
     */
    public static TimeSeries getOrAddSeriesFromDataset(TimeSeriesCollection dataset, Comparable seriesKey) {
      TimeSeries timeSeries = dataset.getSeries(seriesKey);
      if ( timeSeries == null ) {
        timeSeries = new TimeSeries(seriesKey);
        dataset.addSeries( timeSeries );
      }
      return timeSeries;
    }

    /**
	 * Searchs the renderers of a plot for {@link SelectionRenderer} for a
	 * {@link Dataset}.
	 * 
	 * @param plot
	 *            the plot to check for renderer
	 * @param dataset
	 *            the dataset a {@link SelectionRenderer} is searched for
	 * @return {@code null} if there is no {@link SelectionRenderer} defined for
	 *         the dataset
	 */
	public static SelectionRenderer<?> getSelectionRendererForDataset(
			Plot plot, Dataset dataset) {
		if (plot instanceof XYPlot) {
			XYPlot xyPlot = (XYPlot) plot;
			for (int i = 0; i < xyPlot.getDatasetCount(); i++) {
				Dataset d = xyPlot.getDataset(i);
				if (d != dataset)
					continue;
				// determine renderer for dataset (if no renderer is set,
				// take the primary renderer)
				XYItemRenderer r = xyPlot.getRenderer(i);
				if (r == null)
					r = xyPlot.getRenderer();
				// stop search, if renderer is a SelectionRenderer
				if (r instanceof SelectionRenderer<?>)
					return (SelectionRenderer<?>) r;
			}
		}

		if (plot instanceof CategoryPlot) {
			CategoryPlot catPlot = (CategoryPlot) plot;
			for (int i = 0; i < catPlot.getDatasetCount(); i++) {
				Dataset d = catPlot.getDataset(i);
				if (d != dataset)
					continue;
				// determine renderer for dataset (if no renderer is set,
				// take the primary renderer)
				CategoryItemRenderer r = catPlot.getRenderer(i);
				if (r == null)
					r = catPlot.getRenderer();
				// stop search, if renderer is a SelectionRenderer
				if (r instanceof SelectionRenderer<?>)
					return (SelectionRenderer<?>) r;
			}
		}

		return null;
	}

	/**
	 * Searchs the renderers of a plot for {@link SelectionRenderer} for a
	 * {@link Dataset} and returns its {@link DatasetSelectionModel}.
	 * 
	 * @param plot
	 *            the plot to check for renderer
	 * @param dataset
	 *            the dataset a {@link SelectionRenderer} is searched for
	 * @return {@code null} if there is no {@link SelectionRenderer} defined for
	 *         the dataset
	 */
	public static DatasetSelectionModel<?, ?, ?> getSelectionModelForDataset(
			Plot plot, Dataset dataset) {
		SelectionRenderer<?> selRenderer = getSelectionRendererForDataset(plot,
				dataset);
		if (selRenderer != null)
			return selRenderer.getSelectionModel();
		return null;
	}

	/**
	 * Returns all {@link SelectionRenderer} for a {@link Plot}.
	 * 
	 * @param plot
	 *            the plot to check for renderer
	 * @return never {@code null}
	 */
	public static Map<Integer, SelectionRenderer<?>> getSelectionRenderer(
			Plot plot) {
		Map<Integer, SelectionRenderer<?>> selRenderer = new HashMap<Integer, SelectionRenderer<?>>();

		if (plot instanceof XYPlot) {
			XYPlot xyPlot = (XYPlot) plot;
			for (int i = 0; i < xyPlot.getRendererCount(); i++) {
				XYItemRenderer r = xyPlot.getRenderer(i);
				if (r instanceof SelectionRenderer<?>)
					selRenderer.put(i, (SelectionRenderer<?>) r);
			}
		}

		if (plot instanceof CategoryPlot) {
			CategoryPlot catPlot = (CategoryPlot) plot;
			for (int i = 0; i < catPlot.getRendererCount(); i++) {
				CategoryItemRenderer r = catPlot.getRenderer(i);
				if (r instanceof SelectionRenderer<?>)
					selRenderer.put(i, (SelectionRenderer<?>) r);
			}
		}

		return selRenderer;
	}

	/**
	 * Searchs the renderers of a plot for all {@link SelectionRenderer} and
	 * returns their {@link DatasetSelectionModel DatasetSelectionModels}.
	 * 
	 * @param plot
	 *            the plot to check for renderer
	 * @return never {@code null}
	 */
	public static Map<Integer, DatasetSelectionModel<?, ?, ?>> getSelectionModels(
			Plot plot) {
		Map<Integer, SelectionRenderer<?>> selRenderer = getSelectionRenderer(plot);
		Map<Integer, DatasetSelectionModel<?, ?, ?>> selModels = new HashMap<Integer, DatasetSelectionModel<?, ?, ?>>();
		for (int dataset : selRenderer.keySet())
			selModels
					.put(dataset, selRenderer.get(dataset).getSelectionModel());
		return selModels;
	}

	/**
	 * Returns the (first) dataset of a {@link Plot} which is rendererd by a
	 * renderer.
	 * 
	 * @param plot
	 *            the plot which holds the datasets and renderer
	 * @param renderer
	 *            the renderer a dataset is seached for
	 * @return {@code null} if there is no dataset which is rendered by the
	 *         given renderer
	 */
	public static Dataset getDatasetForRenderer(Plot plot, Object renderer) {
		if (plot instanceof XYPlot) {
			XYPlot xyPlot = (XYPlot) plot;
			for (int i = 0; i < xyPlot.getRendererCount(); i++)
				if (xyPlot.getRenderer(i) == renderer)
					return xyPlot.getDataset(i);
		}

		if (plot instanceof CategoryPlot) {
			CategoryPlot catPlot = (CategoryPlot) plot;
			for (int i = 0; i < catPlot.getRendererCount(); i++)
				if (catPlot.getRenderer(i) == renderer)
					return catPlot.getDataset(i);
		}

		return null;
	}

	/**
	 * Applies general dataset properties to a {@link Dataset}.
	 * 
	 * @param dataset
	 *            dataset the properties are applied to
	 * @param sample
	 *            dataset the properties are taken from
	 */
	public static void applyDatasetProperties(Dataset dataset, Dataset sample) {
		dataset.setGroup(sample.getGroup());
	}

	/**
	 * Creates a {@link JFreeChart} which shows a point for each data item and a
	 * the appropriate regression line.
	 * 
	 * @param dataset
	 *            the data
	 * @param title
	 *            title for the chart
	 * @param xAxisTitle
	 *            title for the X-Axis
	 * @param yAxisTitle
	 *            title for the Y-Axis
	 * @param regressionLine
	 *            indicates whether the regression line is shown
	 */
	public static JFreeChart createRegressionChart(XYDataset dataset,
			String title, String xAxisTitle, String yAxisTitle,
			boolean regressionLine, boolean regressionInfo) {
		if (regressionLine && !(dataset instanceof XYSeriesCollection))
			throw new UnsupportedOperationException(
					"Regression line can only be created for XYSeriesCollection!");

		NumberAxis xAxis = new NumberAxis(xAxisTitle);
		xAxis.setAutoRangeIncludesZero(false);
		NumberAxis yAxis = new NumberAxis(yAxisTitle);
		yAxis.setAutoRangeIncludesZero(false);

		XYPlot plot = new XYPlot(dataset, xAxis, yAxis, null);
		XYToolTipGenerator toolTipGenerator = new StandardXYToolTipGenerator();
		XYURLGenerator urlGenerator = new StandardXYURLGenerator();
		XYItemRenderer renderer = new XYLineAndShapeRenderer(false, true);

		renderer.setBaseToolTipGenerator(toolTipGenerator);
		renderer.setURLGenerator(urlGenerator);
		plot.setRenderer(renderer);
		plot.setOrientation(PlotOrientation.VERTICAL);

		// Create Regression line
		if (regressionLine) {

			// only if there is at least one series with at least 2 items, otherwise we would get a
			// IllegalArgumentException, or a " Series index out of bounds" exception
			XYSeriesCollection xySeriesCollection = (XYSeriesCollection) dataset;
			if ((xySeriesCollection.getSeriesCount() > 0) && (xySeriesCollection.getItemCount(0) >= 2)) {
				// create sample data for curve (plot function directly is not
				// yet available)
				RegressionDataset regressionData = createRegressionLineDataset(
						(XYSeriesCollection) dataset, 0, title
								+ " "+RESOURCE.getString("regressionline"), 2);
				if (regressionData != null) {
					addRegressionLineToPlot(plot, regressionData, Color.blue);
					if ( regressionInfo )
					  addRegressionInfoToPlot(plot, regressionData);
				}
			} else {
				LOGGER
						.info("Not enought items in the series to create a regression line.");
			}
		}
		
		JFreeChart chart = new JFreeChart(title, JFreeChart.DEFAULT_TITLE_FONT,
		    plot, true);

		return chart; // ChartFactory.createScatterPlot(title, xAttr, yAttr,
		// dataset, PlotOrientation.HORIZONTAL, true, true,
		// true);
	}

	/**
	 * Creates sample data for a regression line of an {@link XYDataset}.
	 * 
	 * @param dataset
	 *            data the regression line is created for
	 * @param series
	 *            series of {@code dataset} the regression line is created for
	 * @param seriesKey
	 *            the key for the created dataset
	 * @param sampleCnt
	 *            count of created samples
	 */
	public static RegressionDataset createRegressionLineDataset(
			XYSeriesCollection dataset, int series, String seriesKey,
			int sampleCount) {
		
		double minX = Double.MIN_VALUE;
		double maxX = Double.MAX_VALUE;
		
		if (dataset.getSeriesCount() > 0) {
			minX = dataset.getSeries(0).getMinX();
			maxX = dataset.getSeries(0).getMaxX();
		} 
		return createRegressionLineDataset(dataset, series, seriesKey, minX,
				maxX, sampleCount);
	}

	/**
	 * Returns the parameters 'a' and 'b' for an equation y = a + bx, fitted to
	 * the data using ordinary least squares regression. The result is returned
	 * as a double[], where result[0] --> a, and result[1] --> b.
	 * 
	 * @param data
	 *            the data.
	 * @param series
	 *            the series (zero-based index).
	 * 
	 * @return Steigung und Y-Achsenabstand sowie den Produkmomentkorrelationskoefizienten r
	 */
	public static double[] getOLSRegression(XYDataset data, int series) {

		int n = data.getItemCount(series);
		if (n < 2) {
			throw new IllegalArgumentException("Not enough data.");
		}

		double sumX = 0;
		double sumXsq = 0;
		double sumY = 0;
		double sumYsq = 0;
		double sumXX = 0;
		double sumXY = 0;
		for (int i = 0; i < n; i++) {
			double x = data.getXValue(series, i);
			double y = data.getYValue(series, i);
			sumX += x;
			sumXsq += x * x;
			sumY += y;
			sumYsq += y * y;
			double xx = x * x;
			sumXX += xx;
			double xy = x * y;
			sumXY += xy;
		}
		double sxx = sumXX - (sumX * sumX) / n;
		double sxy = sumXY - (sumX * sumY) / n;
		double xbar = sumX / n;
		double ybar = sumY / n;

		double[] result = new double[4];
		result[1] = sxy / sxx;
		result[0] = ybar - result[1] * xbar;
		
		// Added by me:
		result[2] = (n * sumXY - sumX * sumY)
				/ Math.sqrt((n * sumXsq - sumX * sumX)
						* (n * sumYsq - sumY * sumY));

		result[3] = n;
		return result;

	}

	/**
	 * Creates sample data for a regression line of an {@link XYDataset}.
	 * 
	 * @param dataset
	 *            data the regression line is created for
	 * @param series
	 *            series of {@code dataset} the regression line is created for
	 * @param seriesKey
	 *            the key for the created dataset
	 * @param startX
	 *            the first value the sample data is created for
	 * @param endX
	 *            the last value the sample data is created for
	 * @param sampleCnt
	 *            count of created samples
	 * @return {@code null} if the regression data could not be created (because
	 *         of an error)
	 */
	public static RegressionDataset createRegressionLineDataset(XYDataset dataset,
			int series, String seriesKey, double startX, double endX,
			int sampleCount) {
		try {
//			double[] coefficients = getOLSRegression(dataset, series);
//			double r = coefficients[2];
//			String rFormatted = getRFormatted(r);
//			String rQuality   = getRQuality(r);
//			
//			// Put additional legend informations in series key
//			// NOTE: The KECK projects parses this information!!
//			//       So be aware of uncontrolled changes to this
//			//       information (or its format)!!!
//			seriesKey += ", r=" + rFormatted;
//
//			// Depending on the quality of the regression, we add a note
//			// (c) LAUX
//  			seriesKey += " (" + rQuality + ")";
//
//			Function2D curve = new LineFunction2D(coefficients[0],
//					coefficients[1]);
//			
//// SK: Played around to make the line longer, but then the view will automatically get bigger also.. So it is probably OK the way it is			
////			return DatasetUtilities.sampleFunction2D(curve, startX-Math.abs(endX-startX)*10, endX+Math.abs(endX-startX)*10, 2,
////					seriesKey);
//			
//			return DatasetUtilities.sampleFunction2D(curve, startX, endX, sampleCount,
//			seriesKey);
		  return RegressionDatasetImpl.create(dataset, series, seriesKey, startX, endX, sampleCount);
			
		} catch (Exception err) {
			// according to the data sometimes the regression
			// dataset could not be created
			LOGGER.warn("Regression line could not be created: "
					+ err.getLocalizedMessage());
			return null;
		}
	}
	
	/**
	 * Returns a quality representation of the R-value (e.g. weak,
	 * medium, strong, ...) according to LAUX.
	 */
	public static String getRQuality(double r) {
		if (Double.isNaN(r))
			return RESOURCE.getString("regression.notPossible");
			
	    double rAbs = Math.abs(r);
	    // Depending on the quality of the regression, we add a note
	    // (c) LAUX
	    String qualityKey = "";
	    String negativeKeySuffix = r < 0 ? ".neg" : "";
	    
	    
	    if (rAbs < 0.01) {
	      qualityKey = "regression.none";
	    } else if (rAbs < 0.5) {
	      qualityKey = "regression.weak"+negativeKeySuffix;
	    } else if (rAbs < 0.7) {
	      qualityKey = "regression.medium"+negativeKeySuffix;
	    } else if (rAbs < 0.86) {
	      qualityKey = "regression.strong"+negativeKeySuffix;
	    } else if (rAbs < 1.) {
	      qualityKey = "regression.verystrong"+negativeKeySuffix;
	    } else
	      qualityKey = "regression.total"+negativeKeySuffix;

	     return RESOURCE.getString(qualityKey);
	}
	
	/**
	 * Returns a formatted representation of the R-value.
	 */
	public static String getRFormatted(double r) {
	    String rFormatted = DecimalFormat.getInstance().format(r);
	    if (rFormatted.startsWith("0"))
	        rFormatted = rFormatted.substring(1);
	    if (rFormatted.startsWith("-0"))
	        rFormatted = "-"+rFormatted.substring(2);
	     return rFormatted;
	}

	/**
	 * Adds a line plot for regression data to a plot.
	 * 
	 * @param plot
	 *            the plot the line is added to (at the end)
	 * @param regressionData
	 *            the regression data
	 * @param lineColor
	 *            the color for the regression line
	 */
	public static void addRegressionLineToPlot(XYPlot plot,
			RegressionDataset regressionData, Color lineColor) {
		int datasetCount = plot.getDatasetCount();
		// set the regression data
		plot.setDataset(datasetCount, regressionData);
		// create and set the renderer for the regression line
		XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer(true,
				false);
		renderer.setSeriesPaint(0, Color.blue);
		plot.setRenderer(datasetCount, renderer);
	}
	
    /**
     * Adds annotations to a plot which contains textual informations
     * about the regression.
     * 
     * @param plot
     *            the plot the annotation is added to (at top left)
     * @param regressionData
     *            the regression data
     */
	public static void addRegressionInfoToPlot(XYPlot plot, RegressionDataset regressionData) {
      DecimalFormat formatter = new DecimalFormat("0.000");
      
		if (Double.isNaN(regressionData.getR()))
			return;

      // Container which hold the title and text
      BlockContainer container = new BlockContainer( new ColumnArrangement() );
      
      // Information title
      String regrInfoTitleStr = R("regressionInfo.title") + ":";
      TextTitle regrInfoTitle = new TextTitle(regrInfoTitleStr, TextTitle.DEFAULT_FONT.deriveFont(Font.BOLD) );
      regrInfoTitle.setTextAlignment(HorizontalAlignment.LEFT);
      container.add(regrInfoTitle);
      
      // Information text for R
      String regrInfoRStr = R("regressionInfo.data.r");
      regrInfoRStr += " = " + formatter.format(regressionData.getR());
      regrInfoRStr += " (" + regressionData.getRQuality() + ")";
      TextTitle regrInfoR = new TextTitle(regrInfoRStr, TextTitle.DEFAULT_FONT.deriveFont(Font.PLAIN) );
      regrInfoR.setTextAlignment(HorizontalAlignment.LEFT);
      regrInfoR.setToolTipText(R("regressionInfo.data.r.tooltip",regressionData.getR()));
      container.add(regrInfoR);
      
      // Information text for N
      String regrInfoNStr = R("regressionInfo.data.n");
      regrInfoNStr += " = " + regressionData.getSampleCount();
      TextTitle regrInfoN = new TextTitle(regrInfoNStr, TextTitle.DEFAULT_FONT.deriveFont(Font.PLAIN) );
      regrInfoN.setTextAlignment(HorizontalAlignment.LEFT);
      regrInfoN.setToolTipText( R("regressionInfo.data.n.tooltip",regressionData.getSampleCount()) );
      container.add(regrInfoN);

      // Information text for the regression line formular
      String regrInfoFormulaStr = R("regressionInfo.data.formula");
      regrInfoFormulaStr += " = " + formatter.format(regressionData.getRegressionLineSlope()) + "x";
      regrInfoFormulaStr += ( regressionData.getRegressionLineIntercept() >= 0 ? "  + " : "  - " );
      regrInfoFormulaStr += formatter.format(Math.abs(regressionData.getRegressionLineIntercept()));
      TextTitle regrInfoFormula = new TextTitle(regrInfoFormulaStr, TextTitle.DEFAULT_FONT.deriveFont(Font.PLAIN) );
      regrInfoFormula.setTextAlignment(HorizontalAlignment.LEFT);
      regrInfoFormula.setToolTipText(
          R("regressionInfo.data.formula.tooltip",
              regressionData.getRegressionLineSlope(),
              regressionData.getRegressionLineIntercept()
          )
      );
      container.add(regrInfoFormula);

      // Create a Title for the BlockContainer
      CompositeTitle compTitle = new CompositeTitle(container);
      
      // Create the Annotation and add it to the plot
      XYTitleAnnotation regrInfoAnno = null;
      if ( regressionData.getRegressionLineSlope() >= 0 )
        // if slope is positive, the regression line is plot
        // from bottom left to top right, so put the information
        // on the top left site of the plot
        regrInfoAnno = new XYTitleAnnotation(0.02, 0.98, compTitle, RectangleAnchor.TOP_LEFT);
      else
        // if slope is negative, the regression line is plot
        // from top left to bottom right, so put the information
        // on the bottom left site of the plot
        regrInfoAnno = new XYTitleAnnotation(0.02, 0.02, compTitle, RectangleAnchor.BOTTOM_LEFT);
      plot.addAnnotation( regrInfoAnno );
	}

	
	/**
     * Creates a basic stroke by only some basic attributes
     * @param lineWidth line width
     * @param dashAttr  defines the dashing
     * @return
     */
    public static Stroke createDefaultStroke(float lineWidth, Float[] dashAttr) {
    	float[] floats = null;
    	if (dashAttr != null) {
    		floats = new float[dashAttr.length];
    		for (int i=0; i<floats.length; i++)
    			floats[i] = dashAttr[i];
    	}
      return createDefaultStroke(lineWidth, floats);
    }
    
	/**
	 * Creates a basic stroke by only some basic attributes
	 * @param lineWidth line width
	 * @param dashAttr  defines the dashing
	 * @return
	 */
	public static Stroke createDefaultStroke(float lineWidth, float[] dashAttr) {
	  //*** Set the stroke parameters ***
	  float   width      = lineWidth;  // Linien Breite
      int     cap        = BasicStroke.CAP_ROUND; // Rendering am Ende der Linie
      int     join       = BasicStroke.JOIN_ROUND; // Rendering an den Zwischenknoten
      float   miterlimit = 1.0f;
      float[] dash       = dashAttr; // may be null
      float   dash_phase = 1.0f;
      Stroke stroke = new BasicStroke(width, cap, join, miterlimit, dash, dash_phase);
	  
	  return stroke;
	}
	
	// ###################
	// BISHER NUR WILDE TESTS
	// ###################
	// /**
	// * Converts a Excel file to CVS format.
	// * @param xlsInput stream to Excel source
	// * @param sheetName name of the sheet in the Excel source (if {@code null}
	// the
	// * first sheet is used)
	// * @param headRow indicates whether the first row contains the column
	// names
	// * @param csvHeader indicates whether the column names are also taken to
	// the CVS (note:
	// * if {@code true} and {@code headRow = false} dummy column
	// * names "A", "B", ... are generates in CVS!)
	// * @param delimChr character takes as delimiter in CSV
	// * @param strEncChr character String values are encapsulated with (if
	// {@code null}, Strings
	// * are not encapsulated)
	// */
	// public static InputStream createCSVFromXLS(InputStream xlsInput, String
	// sheetName, boolean headRow, boolean csvHeader, String delimChr, String
	// strEncChar) throws IOException {
	// PipedOutputStream csvOutput = new PipedOutputStream();
	//   
	// // read XLS and create a CSV
	// HSSFWorkbook wb = new HSSFWorkbook( new POIFSFileSystem(xlsInput) );
	// HSSFSheet sheet = ( sheetName != null ) ? wb.getSheet(sheetName) :
	// wb.getSheetAt(0);
	//    
	// int firstRow = sheet.getFirstRowNum();
	// int lastRow = sheet.getLastRowNum();
	// // int rowCount = sheet.getPhysicalNumberOfRows();
	// for (int y=firstRow; y <= lastRow; y++) {
	// HSSFRow row = sheet.getRow(y);
	// int firstCol = row.getFirstCellNum();
	// int lastCol = row.getLastCellNum();
	// for (int x=firstCol; x <= lastCol; x++) {
	// HSSFCell cell = row.getCell(x);
	// // String csvValue = convertCellValueToString(cell,???);
	// }
	// }
	//   
	// return new PipedInputStream(csvOutput);
	// }
	//  
	// public static String convertCellValueToString(HSSFCell cell,
	// HSSFFormulaEvaluator formulaEvaluator) {
	// String strValue = "";
	// switch( cell.getCellType() ) {
	// case HSSFCell.CELL_TYPE_BOOLEAN: strValue =
	// ""+cell.getBooleanCellValue();
	// break;
	// case HSSFCell.CELL_TYPE_NUMERIC: strValue =
	// ""+cell.getNumericCellValue();
	// break;
	// case HSSFCell.CELL_TYPE_FORMULA: strValue =
	// ""+cell.getNumericCellValue();
	// }
	// return strValue;
	//    
	// }
	//
	// /**
	// * Creates a Dataset from Excel file.
	// * @param xlsInput stream to Excel source
	// * @param sheetName name of the sheet in the Excel source (if {@code null}
	// the
	// * first sheet is used)
	// */
	// public static CategoryDataset createDatasetFromXLS(InputStream xlsInput,
	// String sheetName) throws IOException {
	// PipedOutputStream csvOutput = new PipedOutputStream();
	// InputStreamReader csvInput = new InputStreamReader( new
	// PipedInputStream(csvOutput) );
	//    
	//
	// // return new CSV(';','"').readCategoryDataset(csvInput);
	//
	// HSSFWorkbook wb = new HSSFWorkbook(new POIFSFileSystem(xlsInput));
	// ExcelExtractor extractor = new ExcelExtractor(wb);
	//    
	// extractor.setFormulasNotResults(false);
	// extractor.setIncludeSheetNames(false);
	// String text = extractor.getText();
	// System.out.println(text);
	// return null;
	// }
}
