package de.schmitzm.jfree.chart;

import org.jfree.chart.LegendItem;
import org.jfree.chart.LegendItemCollection;
import org.jfree.chart.LegendItemSource;

import de.schmitzm.jfree.JFreeChartUtil;

/**
 * This {@link LegendItemSource} tunnels an other {@link LegendItemSource} or
 * a {@link LegendItemCollection} and filters out some {@link LegendItem LegendItems}
 * by the prefix of their {@linkplain LegendItem#getLabel() labels}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class FilteredLegendItemSource implements LegendItemSource {
  /** Holds the filtered legend items. */
  protected LegendItemCollection filteredLegendItems = null;
  /** Holds the prefix the original legend items are filtered with. */
  protected String filterPrefix = null;
  
  /**
   * Creates a new legend item source.
   * @param legendItems   original legend items
   * @param filterPrefix  all items with this prefix are filtered out (can be {@code null})
   */
  public FilteredLegendItemSource(LegendItemCollection legendItems, String filterPrefix) {
    this.filterPrefix        = filterPrefix;
    this.filteredLegendItems = filterLegendItems(legendItems,filterPrefix);
  }
  
  /**
   * Creates a new legend item source. {@link JFreeChartUtil#DUMMY_SERIES_PREFIX} is
   * used as filter prefix.
   * @param legendItems   original legend items 
   */
  public FilteredLegendItemSource(LegendItemCollection legendItems) {
    this(legendItems,JFreeChartUtil.DUMMY_SERIES_PREFIX);
  }

  /**
   * Creates a new legend item source. {@link JFreeChartUtil#DUMMY_SERIES_PREFIX} is
   * used as filter prefix.
   * @param lis holds the original legend items 
   * @param filterPrefix  all items with this prefix are filtered out (can be {@code null})
   */
  public FilteredLegendItemSource(LegendItemSource lis, String filterPrefix) {
    this(lis.getLegendItems(), filterPrefix);
  }

  /**
   * Creates a new legend item source. {@link JFreeChartUtil#DUMMY_SERIES_PREFIX} is
   * used as filter prefix.
   * @param lis holds the original legend items 
   */
  public FilteredLegendItemSource(LegendItemSource lis) {
    this(lis.getLegendItems());
  }
  
  /**
   * Returns the filtered legend items.  
   */
  @Override
  public LegendItemCollection getLegendItems() {
    return filteredLegendItems;
  }

  /**
   * Filters the items of a {@link LegendItemCollection} by a prefix string
   * of their labels.
   * @param legendItems   original legend items
   * @param filterPrefix  all items with this prefix are filtered out (can be {@code null})
   * @return the original {@link LegendItemCollection} if {@code filterPrefix} is {@code null}
   * @see LegendItem#getLabel()
   */
  public static LegendItemCollection filterLegendItems(LegendItemCollection legendItems, String filterPrefix) {
    if ( legendItems == null || filterPrefix == null )
      return legendItems;
    
    LegendItemCollection filteredLegendItems = new LegendItemCollection();
    for (int j=0; j<legendItems.getItemCount(); j++) {
      LegendItem lItem = legendItems.get(j);
      if (!lItem.getLabel().startsWith(filterPrefix))
        filteredLegendItems.add(lItem);
    }
    
    return filteredLegendItems;
  }
}