/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package de.schmitzm.jfree.chart;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.labels.CategoryToolTipGenerator;
import org.jfree.chart.labels.PieToolTipGenerator;
import org.jfree.chart.labels.XYToolTipGenerator;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.category.AbstractCategoryItemRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.PieDataset;
import org.jfree.data.xy.XYDataset;

import de.schmitzm.jfree.JFreeChartUtil;
import de.schmitzm.jfree.chart.style.ChartStyle;

/**
 * A tooltip generator for data items, which uses the axis and series legend label in
 * the tooltip, instead of the simple "series key", as the standard
 * tooltip generators do.<br>
 * Furthermore this tooltip generator can be used for XY-datasets as well
 * as for category-datasets and pie-datasets. 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *
 */
public class StyledToolTipGenerator implements CategoryToolTipGenerator,
    XYToolTipGenerator, PieToolTipGenerator {

  /** Holds the chart to access the axis and legend. */
  protected JFreeChart chart;
  /** Holds the chart style to determine axis and legend labels directly
   *  from the style (currently not used!). */
  protected ChartStyle chartStyle;
  
  /**
   * Creates a new tooltip generator.
   * @param chart  the chart to access the axis and legend
   * @param style  the chart style to determine axis and legend labels directly
   *               from the style (currently not used!)
   */
  public StyledToolTipGenerator(JFreeChart chart, ChartStyle style) {
    this.chart      = chart;
    this.chartStyle = style;
  }
  
  /**
   * Combines the axis label, domain value, series key and range value to
   * a complete tooltip string.
   * @param axisLabel   the axis description
   * @param domainValue the domain value of the data item
   * @param seriesKey   the series key of the data item
   * @param rangeValue  the range value of the data item
   */
  private String generateToolTip(String axisLabel, Object domainValue, Object seriesKey, Object rangeValue) {
    StringBuffer sb = new StringBuffer();
    // Append Range Axis label
    if ( axisLabel != null && !"".equals(axisLabel.trim()) )
      sb.append(axisLabel.trim());
    // Append domain value (X or Category)
    if ( domainValue != null && !"".equals(domainValue.toString().trim()) ) {
      if ( sb.length() > 0 )
        sb.append(", ");
      sb.append(domainValue.toString().trim());
    }
    // Append series
    if ( seriesKey != null && !"".equals(seriesKey.toString().trim()) ) {
      if ( sb.length() > 0 )
        sb.append(", ");
      sb.append(seriesKey.toString().trim());
    }
    // Append range value
    if ( sb.length() > 0 )
      sb.append(": ");
    sb.append(rangeValue);
    
    return sb.toString();

  }
  
  
  /**
   * Generates a tooltip for a data item.
   * @param dataset the dataset
   * @param row     the row index (zero-based)
   * @param col     the column index (zero-based)
   */
  @Override
  public String generateToolTip(CategoryDataset dataset, int row, int col) {
    CategoryPlot plot = chart.getCategoryPlot();
    AbstractCategoryItemRenderer rend = (AbstractCategoryItemRenderer)plot.getRendererForDataset(dataset);
    ValueAxis rangeAxis = JFreeChartUtil.getRangeAxisForDataset(plot, dataset);  

    String rangeAxisLabel = rangeAxis.getLabel();
    Object domainValue    = dataset.getColumnKey(col);
    String seriesValue    = rend.getLegendItemLabelGenerator().generateLabel(dataset, row);
    Object rangeValue     = dataset.getValue(row, col);
    
    return generateToolTip(rangeAxisLabel,domainValue,seriesValue,rangeValue);
  }

  /**
   * Generates a tooltip for a data item.
   * @param dataset the dataset
   * @param series  the series index (zero-based)
   * @param item    the item index (zero-based)   
   */
  @Override
  public String generateToolTip(XYDataset dataset, int series, int item) {
    XYPlot plot = chart.getXYPlot();
    XYItemRenderer rend = (XYItemRenderer)plot.getRendererForDataset(dataset);
    ValueAxis rangeAxis = JFreeChartUtil.getRangeAxisForDataset(plot, dataset);  

    String rangeAxisLabel = rangeAxis.getLabel();
    Object domainValue    = dataset.getX(series, item);
    String seriesValue    = rend.getLegendItemLabelGenerator().generateLabel(dataset, series);
    Object rangeValue     = dataset.getY(series, item);
    
    return generateToolTip(rangeAxisLabel,domainValue,seriesValue,rangeValue);
  }

  /**
   * Generates a tooltip for a data item.
   * @param dataset the dataset
   * @param key     the pie data key
   */
  @Override
  public String generateToolTip(PieDataset dataset, Comparable key) {
    throw new UnsupportedOperationException("StyledToolTipGenerator.generateToolTip(PieDataset,Comparable) not yet implemented!");
  }

}
