/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/
package de.schmitzm.jfree.chart.style;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Shape;
import java.awt.Stroke;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Category;
import org.jfree.chart.labels.CategorySeriesLabelGenerator;
import org.jfree.chart.labels.StandardCategorySeriesLabelGenerator;
import org.jfree.chart.labels.StandardXYSeriesLabelGenerator;
import org.jfree.chart.labels.XYSeriesLabelGenerator;
import org.jfree.chart.renderer.AbstractRenderer;
import org.jfree.chart.renderer.category.AbstractCategoryItemRenderer;
import org.jfree.chart.renderer.xy.AbstractXYItemRenderer;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;

import de.schmitzm.data.Copyable;
import de.schmitzm.jfree.JFreeChartUtil;
import de.schmitzm.lang.LangUtil;

/**
 * This class defines the rendering style the charts series. The following
 * properties are available:
 * <ul>
 *   <li>color of series graphs</li>
 *   <li>line width of series graphs</li>
 *   <li>dashing attributes for the series graphs</li>
 *   <li>visibility of series item labels</li>
 *   <li>visibility of series shapes</li>
 *   <li>the legend labels of the series</li>
 *   <li>visibility of legend label</li>
 *   <li>percentage margin between bars (only applicable for {@link XYBarRenderer})</li>
 * </ul>

 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class ChartRendererStyle implements Copyable<ChartRendererStyle> {
  /** Logger for this class */
  protected final Category LOGGER = LangUtil.createLogger(this);

  /** Standard label generator used for XY-Charts. */
  protected XYSeriesLabelGenerator LABELGEN_XY = new StandardXYSeriesLabelGenerator();
  /** Standard label generator used for Category-Charts. */
  protected CategorySeriesLabelGenerator LABELGEN_CAT = new StandardCategorySeriesLabelGenerator();

  /** Holds whether the item labels are visible (for all series this property
   *  is not set explicitly). */
  protected Boolean defaultItemLabelsVisible = null;
  /** Holds the color a series is painted with (for all series this property
   *  is not set explicitly). */
  protected Color defaultPaint = null;
  /** Holds the line width a series graph is painted with (for all series this property
   *  is not set explicitly). */
  protected Integer defaultWidth = null;
  /** Holds the dash attributes a series graph is painted with (for all series this property
   *  is not set explicitly).
   *  @see BasicStroke */
  protected Float[] defaultDashAttr = null;
  /** Holds whether the shape of a series is visible (for all series this property
   *  is not set explicitly). */
  protected Boolean defaultShapesVisible = null;
  /** Holds whether the series is visible in the legend (for all series this property
   *  is not set explicitly). */
  protected Boolean defaultLegendVisible = null;
  /** Holds the label shown in the legend for the series (for all series this property
   *  is not set explicitly). */
  protected ChartLabelStyle defaultLegendLabel = null;
  /** Holds the tooltip shown in the legend for the series (for all series this property
   *  is not set explicitly). */
  protected ChartLabelStyle defaultLegendTooltip = null;

  /** Holds the ID of the series (series key of the dataset) the series style definition has
   *  to be used for.*/
  protected Map<Integer, String> seriesKey = new HashMap<Integer, String>();
  /** Holds whether the item labels are visible for a series. */
  protected Map<Integer, Boolean> seriesItemLabelsVisible = new HashMap<Integer, Boolean>();
  /** Holds the color a series is painted with. */
  protected Map<Integer, Color> seriesPaint = new HashMap<Integer, Color>();
  /** Holds the line width a series graph is painted with. */
  protected Map<Integer, Integer> seriesWidth = new HashMap<Integer, Integer>();
  /** Holds the dash attributes a series graph is painted with.
   *  @see BasicStroke */
  protected Map<Integer,Float[]> seriesDashAttr = new HashMap<Integer,Float[]>();
  /** Holds whether the shape of a series is visible. */
  protected Map<Integer, Boolean> seriesShapesVisible = new HashMap<Integer, Boolean>();
  /** Holds whether the series is visible in the legend. */
  protected Map<Integer, Boolean> seriesLegendVisible = new HashMap<Integer, Boolean>();
  /** Holds the label shown in the legend for the series. */
  protected Map<Integer, ChartLabelStyle> seriesLegendLabel = new HashMap<Integer, ChartLabelStyle>();
  /** Holds the tooltip shown in the legend for the series. */
  protected Map<Integer, ChartLabelStyle> seriesLegendTooltip = new HashMap<Integer, ChartLabelStyle>();
  /** Holds the maximum series index for which a series is specified. */
  protected int maxSeriesIdx = 0;
  /** Holds the (percentage) margin for the bars. Only applicable for {@link XYBarRenderer}. */
  protected double margin = 0.0;
  
  /** Label generator bases on the legend label definitions of this
   *  style. */
  protected LegendLabelGenerator legendLabelGenerator = null;
  /** Label generator bases on the legend tooltip definitions of this
   *  style. */
  protected LegendTooltipGenerator legendTooltipGenerator = null;
  
  /**
   * Creates a new style with default values.
   */
  public ChartRendererStyle() {
    legendLabelGenerator   = new LegendLabelGenerator();
    legendTooltipGenerator = new LegendTooltipGenerator();
  }

  /**
   * Creates a (deep) clone of this style.
   */
  @Override
  public ChartRendererStyle copy() {
    return copyTo( new ChartRendererStyle() );
  }
  
  /**
   * Copies all properties of this style to another one.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   */
  @Override
  public ChartRendererStyle copyTo(ChartRendererStyle dest) {
    if ( dest == null )
      return copy();

    dest.setMargin( getMargin() );
    dest.setDefaultItemLabelsVisible(isDefaultItemLabelsVisible() );
    dest.setDefaultPaint(getDefaultPaint() );
    dest.setDefaultLineWidth(getDefaultLineWidth() );
    dest.setDefaultLineDashAttributes(LangUtil.cloneArray(getDefaultLineDashAttibutes()));
    dest.setDefaultShapesVisible(isDefaultShapesVisible() );
    dest.setDefaultLegendVisible(isDefaultLegendVisible() );
    if ( getDefaultLegendLabel() != null )
      dest.setDefaultLegendLabel(getDefaultLegendLabel().copy() );
    if ( getDefaultLegendTooltip() != null )
      dest.setDefaultLegendTooltip(getDefaultLegendTooltip().copy() );
    
    dest.seriesKey.clear();
    for ( int series : seriesKey.keySet() )
      dest.setSeriesKey( series, getSeriesKey(series) );

    dest.seriesPaint.clear();
    for ( int series : seriesPaint.keySet() )
      dest.setSeriesPaint( series, getSeriesPaint(series) );

    dest.seriesWidth.clear();
    for ( int series : seriesWidth.keySet() )
      dest.setSeriesLineWidth( series, getSeriesLineWidth(series) );

    dest.seriesDashAttr.clear();
    for ( int series : seriesDashAttr.keySet() )
      dest.setSeriesLineDashAttributes( series, LangUtil.cloneArray(getSeriesLineDashAttibutes(series)) );
    
    dest.seriesItemLabelsVisible.clear();
    for ( int series : seriesItemLabelsVisible.keySet() )
      dest.setSeriesItemLabelsVisible( series, isSeriesItemLabelsVisible(series) );

    dest.seriesShapesVisible.clear();
    for ( int series : seriesShapesVisible.keySet() )
      dest.setSeriesShapesVisible( series, isSeriesShapesVisible(series) );

    dest.seriesLegendVisible.clear();
    for ( int series : seriesLegendVisible.keySet() )
      dest.setSeriesLegendVisible( series, isSeriesLegendVisible(series) );

    dest.seriesLegendLabel.clear();
    for ( int series : seriesLegendLabel.keySet() )
      if ( getSeriesLegendLabel(series) != null )
        dest.setSeriesLegendLabel( series, getSeriesLegendLabel(series).copy() );

    dest.seriesLegendTooltip.clear();
    for ( int series : seriesLegendTooltip.keySet() )
      if ( getSeriesLegendTooltip(series) != null )
        dest.setSeriesLegendTooltip( series, getSeriesLegendTooltip(series).copy() );

    return dest;
  }

  /**
   * Returns the margin (percentage) between the chart bars. The margin is
   * only applicable on bar charts.
   */
  public Double getMargin() {
    return margin;
  }

  /**
   * Sets the margin (percentage) between the chart bars. The margin is
   * only applicable on bar charts.
   */
  public void setMargin(Double margin) {
    this.margin = margin;
  }

  /**
   * Returns the count of series the style is defined for.
   */
  public int getSeriesCount() {
    return maxSeriesIdx + 1;
  }
  
  /**
   * Returns the series key (of the dataset), which identifies the series
   * for which the i-th series style is used for.
   * If not set the i-th series style is used for the i-th series. 
   * @param i series index
   * @return {@code null} if no specific color is set for the series
   */
  public String getSeriesKey(int i) {
    return seriesKey.get(i);
  }

  /**
   * Sets the series key (of the dataset), which identifies the series
   * for which the i-th series style is used for.
   * If not set the i-th series style is used for the i-th series.
   * @param i series index
   * @param key the series key in the dataset
   */
  public void setSeriesKey(int i, String key) {
    seriesKey.put(i,key);
    updateSeriesCount();
  }

  /**
   * Returns the color, a series is painted with.
   * @param series series index
   * @return {@code null} if no specific color is set for the series
   */
  public Color getSeriesPaint(int series) {
    return seriesPaint.get(series);
  }

  /**
   * Sets the color, a series is painted with.
   * @param series series index
   * @param color  color for the series (can be {@code null} to reset
   *               to a non-specific color)
   */
  public void setSeriesPaint(int series, Color color) {
    seriesPaint.put(series,color);
    updateSeriesCount();
  }

  /**
   * Returns the line width, a series is painted with.
   * @param series series index
   * @return {@code null} if no specific width is set for the series
   */
  public Integer getSeriesLineWidth(int series) {
    return seriesWidth.get(series);
  }

  /**
   * Sets the line width, a series is painted with.
   * @param series series index
   * @param width  line width for the series (can be {@code null} to reset
   *               to a non-specific color)
   */
  public void setSeriesLineWidth(int series, Integer width) {
    seriesWidth.put(series,width);
    updateSeriesCount();
  }

  /**
   * Returns the dash attributes, a series is painted with.
   * @param series series index
   * @return {@code null} if no specific dashing is set for the series
   * @see BasicStroke
   */
  public Float[] getSeriesLineDashAttibutes(int series) {
    return seriesDashAttr.get(series);
  }

  /**
   * Sets the dash attributes, a series is painted with.
   * @param series series index
   * @param width  dash attributes for the series (can be {@code null} to reset
   *               to a continuous line)
   */
  public void setSeriesLineDashAttributes(int series, Float... dashAttr) {
    if ( dashAttr.length == 0 )
      dashAttr = null;
    seriesDashAttr.put(series,dashAttr);
    updateSeriesCount();
  }

  /**
   * Returns whether the item labels of a series are visible.
   * @param series series index
   * @return {@code null} if property is not set for the series
   */
  public Boolean isSeriesItemLabelsVisible(int series) {
    return seriesItemLabelsVisible.get(series);
  }

  /**
   * Sets whether the item labels of a series are visible.
   * @param series  series index
   * @param visible indicates the visibility of the series (can be
   *                {@code null} to reset to a non-specific property)
   */
  public void setSeriesItemLabelsVisible(int series, Boolean visible) {
    seriesItemLabelsVisible.put(series,visible);
    updateSeriesCount();
  }

  /**
   * Returns whether the shapes of a series are visible.
   * @param series series index
   * @return {@code null} if property is not set for the series
   */
  public Boolean isSeriesShapesVisible(int series) {
    return seriesShapesVisible.get(series);
  }

  /**
   * Sets whether the shapes of a series are visible.
   * @param series  series index
   * @param visible indicates the visibility of the series (can be
   *                {@code null} to reset to a non-specific property)
   */
  public void setSeriesShapesVisible(int series, Boolean visible) {
    seriesShapesVisible.put(series,visible);
    updateSeriesCount();
  }
  
  /**
   * Returns whether the series is shown in the legend.
   * @param series series index
   * @return {@code null} if property is not set for the series
   */
  public Boolean isSeriesLegendVisible(int series) {
    return seriesLegendVisible.get(series);
  }

  /**
   * Sets whether the series is shown in the legend.
   * @param series  series index
   * @param visible indicates the visibility of the series legend item (can be
   *                {@code null} to reset to a non-specific property)
   */
  public void setSeriesLegendVisible(int series, Boolean visible) {
    seriesLegendVisible.put(series,visible);
    updateSeriesCount();
  }
  
  /**
   * Returns the style of the series label in the legend.
   * @param series series index
   * @return {@code null} if property is not set for the series
   */
  public ChartLabelStyle getSeriesLegendLabel(int series) {
    return seriesLegendLabel.get(series);
  }

  /**
   * Sets the style of the series label in the legend.
   * @param series  series index
   * @param labelStyle the style for the series label in the legend (can be
   *                   {@code null} to reset to a non-specific property)
   */
  public void setSeriesLegendLabel(int series, ChartLabelStyle labelStyle) {
    seriesLegendLabel.put(series,labelStyle);
    updateSeriesCount();
  }

  /**
   * Returns the style of the series tooltip in the legend.
   * @param series series index
   * @return {@code null} if property is not set for the series
   */
  public ChartLabelStyle getSeriesLegendTooltip(int series) {
    return seriesLegendTooltip.get(series);
  }

  /**
   * Sets the style of the series tooltip in the legend.
   * @param series  series index
   * @param labelStyle the style for the series tooltip in the legend (can be
   *                   {@code null} to reset to a non-specific property)
   */
  public void setSeriesLegendTooltip(int series, ChartLabelStyle tooltipStyle) {
    seriesLegendTooltip.put(series,tooltipStyle);
    updateSeriesCount();
  }
  
  /**
   * Returns the color, a series is painted with for these
   * series this property is not set explicitly.
   * @return {@code null} if no specific color is set for the series
   */
  public Color getDefaultPaint() {
    return defaultPaint;
  }

  /**
   * Sets the color, a series is painted with for these
   * series this property is not set explicitly.
   * @param color  color for the series (can be {@code null} to reset
   *               to a non-specific color)
   */
  public void setDefaultPaint(Color color) {
    defaultPaint = color;
  }

  /**
   * Returns the line width, a series is painted with for these
   * series this property is not set explicitly.
   * @return {@code null} if no specific width is set for the series
   */
  public Integer getDefaultLineWidth() {
    return defaultWidth;
  }

  /**
   * Sets the line width, a series is painted with for these
   * series this property is not set explicitly.
   * @param width  line width for the series (can be {@code null} to reset
   *               to a non-specific width)
   */
  public void setDefaultLineWidth(Integer width) {
    defaultWidth = width;
  }

  /**
   * Returns the dash attributes, a series is painted with for these
   * series this property is not set explicitly.
   * @return {@code null} if no specific dashing is set for the series
   * @see BasicStroke
   */
  public Float[] getDefaultLineDashAttibutes() {
    return defaultDashAttr;
  }

  /**
   * Sets the dash attributes, a series is painted with for these
   * series this property is not set explicitly.
   * @param width  dash attributes for the series (can be {@code null} to reset
   *               to a continuous line)
   */
  public void setDefaultLineDashAttributes(Float... dashAttr) {
    if ( dashAttr != null && dashAttr.length == 0 )
      dashAttr = null;
    defaultDashAttr = dashAttr;
  }

  /**
   * Returns whether the item labels of a series are visible for these
   * series this property is not set explicitly.
   * @return {@code null} if property is not set for the series
   */
  public Boolean isDefaultItemLabelsVisible() {
    return defaultItemLabelsVisible;
  }

  /**
   * Sets whether the item labels of a series are visible for these
   * series this property is not set explicitly.
   * @param visible indicates the visibility of the series (can be
   *                {@code null} to reset to a non-specific property)
   */
  public void setDefaultItemLabelsVisible(Boolean visible) {
    defaultItemLabelsVisible = visible;
  }

  /**
   * Returns whether the shapes of a series are visible for these
   * series this property is not set explicitly.
   * @return {@code null} if property is not set for the series
   */
  public Boolean isDefaultShapesVisible() {
    return defaultShapesVisible;
  }

  /**
   * Sets whether the shapes of a series are visible for these
   * series this property is not set explicitly.
   * @param visible indicates the visibility of the series (can be
   *                {@code null} to reset to a non-specific property)
   */
  public void setDefaultShapesVisible(Boolean visible) {
    defaultShapesVisible = visible;
  }
  
  /**
   * Returns whether the series is shown in the legend for these
   * series this property is not set explicitly.
   * @return {@code null} if property is not set for the series
   */
  public Boolean isDefaultLegendVisible() {
    return defaultLegendVisible;
  }

  /**
   * Sets whether the series is shown in the legend for these
   * series this property is not set explicitly.
   * @param visible indicates the visibility of the series legend item (can be
   *                {@code null} to reset to a non-specific property)
   */
  public void setDefaultLegendVisible(Boolean visible) {
    defaultLegendVisible = visible;
  }
  
  /**
   * Returns the style of the series label in the legend for these
   * series this property is not set explicitly.
   * @return {@code null} if property is not set for the series
   */
  public ChartLabelStyle getDefaultLegendLabel() {
    return defaultLegendLabel;
  }

  /**
   * Sets the style of the series label in the legend for these
   * series this property is not set explicitly.
   * @param labelStyle the style for the series label in the legend (can be
   *                   {@code null} to reset to a non-specific property)
   */
  public void setDefaultLegendLabel(ChartLabelStyle labelStyle) {
    defaultLegendLabel = labelStyle;
  }

  /**
   * Returns the style of the series tooltip in the legend for these
   * series this property is not set explicitly.
   * @return {@code null} if property is not set for the series
   */
  public ChartLabelStyle getDefaultLegendTooltip() {
    return defaultLegendTooltip;
  }

  /**
   * Sets the style of the series tooltip in the legend for these
   * series this property is not set explicitly.
   * @param labelStyle the style for the series tooltip in the legend (can be
   *                   {@code null} to reset to a non-specific property)
   */
  public void setDefaultLegendTooltip(ChartLabelStyle tooltipStyle) {
    defaultLegendTooltip = tooltipStyle;
  }

  /**
   * Returns a label generator based on the <b>legend label</b>
   * definitions of this style.
   */
  public LegendLabelGenerator getLegendLabelGenerator() {
    return legendLabelGenerator;
  }

  /**
   * Returns a label generator based on the <b>legend tooltip</b>
   * definitions of this style.
   */
  public LegendTooltipGenerator getLegendTooltipGenerator() {
    return legendTooltipGenerator;
  }

  /**
   * Updates the local variable for the defined series with
   * the maximum index (key) of the several hash maps. 
   */
  private void updateSeriesCount() {
    maxSeriesIdx = LangUtil.max(
                      seriesKey.keySet(),
                      seriesPaint.keySet(),
                      seriesWidth.keySet(),
                      seriesDashAttr.keySet(),
                      seriesShapesVisible.keySet(),
                      seriesItemLabelsVisible.keySet(),
                      seriesLegendVisible.keySet(),
                      seriesLegendLabel.keySet(),
                      seriesLegendTooltip.keySet()
    );
  }
  
  /**
   * Applies the style to a specific renderer.<br>
   * <b>Note</b>:
   * <ul>
   *   <li>Some style properties can only be applied to specific renderer types.</li>
   *   <li>The legend properties can not be applied by this method because the
   *       chart / legend is unknown at this point. The properties are set by
   *       {@link ChartStyle#applyToChart(org.jfree.chart.JFreeChart)}</li>
   * </ul>
   * @param renderer a renderer
   */
  public void applyToRenderer(AbstractRenderer renderer) {
    if (renderer == null) {
      LOGGER.debug("No renderer to be customised.");
      return;
    }
      
//    // Apply item label visibility
//    for (int series : seriesItemLabelsVisible.keySet())
//      try {
//        if ( isSeriesItemLabelsVisible(series) != null )
//          renderer.setSeriesItemLabelsVisible(series, isSeriesItemLabelsVisible(series), false);
//      } catch (Exception err) {
//        LOGGER.warn("Renderer contains not series "+series+". Label visibility can not be applied.");
//      }
//    
//    // Apply shape visibility
//    for (int series : seriesShapesVisible.keySet())
//      try {
//        if ( isSeriesShapesVisible(series) != null ) {
//			final Shape shape = isSeriesShapesVisible(series) ? 
//			  AbstractRenderer.DEFAULT_SHAPE
//			  : null;
//			renderer.setSeriesShape(series,
//					  shape,
//					  false);
////			LOGGER.debug("Setting shape = "+shape +" for series "+series);
//		}
//        
//      } catch (Exception err) {
//        LOGGER.warn("Renderer contains not series "+series+". Shapes visibility can not be applied.");
//      }
//
//    // Apply rendering color
//    for (int series : seriesPaint.keySet())
//      try {
//        if ( getSeriesPaint(series) != null )
//          renderer.setSeriesPaint(series, getSeriesPaint(series), false);
//      } catch (Exception err) {
//        LOGGER.warn("Renderer contains not series "+series+". Paint can not be applied.");
//      }
//
//    // Apply legend visibility
//    for (int series : seriesLegendVisible.keySet())
//      try {
//        if ( isSeriesLegendVisible(series) != null )
//          renderer.setSeriesVisibleInLegend(series, isSeriesLegendVisible(series), false);
//      } catch (Exception err) {
//        LOGGER.warn("Renderer contains not series "+series+". Legend visibility can not be applied.");
//      }


    // First apply general default properties to all series of
    // the dataset
    // NOTE: simply set the base-properties does not work (only the first
    //       series is applied by the base-properties! why?)
    Dataset dataset = JFreeChartUtil.getDatasetFrom(renderer);
    int seriesCount = JFreeChartUtil.getSeriesCountFromDataset(dataset);
    for (int series=0; series<seriesCount; series++) { 
      if ( isDefaultItemLabelsVisible() != null )
        renderer.setSeriesItemLabelsVisible(series, isDefaultItemLabelsVisible(), false);
      if ( isDefaultShapesVisible() != null ) {
        Shape shape = isDefaultShapesVisible() ? AbstractRenderer.DEFAULT_SHAPE : null;
        renderer.setSeriesShape(series, shape, false);
      }
      if ( getDefaultPaint() != null )
        renderer.setSeriesPaint(series, getDefaultPaint(), false);
      if ( getDefaultLineWidth() != null || getDefaultLineDashAttibutes() != null ) {
        float   width    = getDefaultLineWidth() == null ? 1.0f : getDefaultLineWidth();
        Float[] dashAttr = getDefaultLineDashAttibutes();
        Stroke stroke = JFreeChartUtil.createDefaultStroke(width, dashAttr);
        renderer.setSeriesStroke(series, stroke, false);
      }
      if ( isDefaultLegendVisible() != null )
        renderer.setSeriesVisibleInLegend(series, isDefaultLegendVisible(), false);
    }    
    
    // Apply explicit series properties
    seriesCount = getSeriesCount();
    for (int i=0; i<seriesCount; i++) {
      // Default: i-th series style for the i-th series
      int series = i;
      // If series key is set, determine the series index in the
      // dataset
      String seriesKey = getSeriesKey(i);
      if ( seriesKey != null ) {
        int idx = JFreeChartUtil.getSeriesIndexFromDataset(dataset, seriesKey);
        if ( idx >= 0 )
          series = idx;
        else {
          LOGGER.warn("Series does not exist in renderer: "+seriesKey+" ("+series+")");
          continue;
        }
      }

      // Apply item label visibility
      if ( isSeriesItemLabelsVisible(series) != null )
        renderer.setSeriesItemLabelsVisible(series, isSeriesItemLabelsVisible(series), false);
      // Apply shape visibility
      if ( isSeriesShapesVisible(series) != null ) {
        Shape shape = isSeriesShapesVisible(series) ? AbstractRenderer.DEFAULT_SHAPE : null;
        renderer.setSeriesShape(series, shape, false);
      }
      // Apply rendering color
      if ( getSeriesPaint(series) != null )
        renderer.setSeriesPaint(series, getSeriesPaint(series), false);
      // Apply rendering stroke (line width) 
      if ( getSeriesLineWidth(series) != null || getSeriesLineDashAttibutes(series) != null ) {
        float   width    = getDefaultLineWidth() == null ? 1.0f : getDefaultLineWidth();
        Float[] dashAttr = getDefaultLineDashAttibutes();
        if ( getSeriesLineWidth(series) != null )
          width = getSeriesLineWidth(series);
        if ( getSeriesLineDashAttibutes(series) != null )
          dashAttr = getSeriesLineDashAttibutes(series);
        Stroke stroke = JFreeChartUtil.createDefaultStroke(width, dashAttr);
        renderer.setSeriesStroke(series, stroke, false);
      }
      // Apply legend visibility
      if ( isSeriesLegendVisible(series) != null )
        renderer.setSeriesVisibleInLegend(series, isSeriesLegendVisible(series), false);
    } 

    
    // Other properties must be applied according to the specific renderer type
    if (renderer instanceof XYBarRenderer)
      applyToBarRenderer((XYBarRenderer)renderer);
    if (renderer instanceof AbstractXYItemRenderer)
      applyToXYItemRenderer((AbstractXYItemRenderer)renderer);
    if (renderer instanceof AbstractCategoryItemRenderer)
      applyToCategoryItemRenderer((AbstractCategoryItemRenderer)renderer);
    
    // notify the listeners about style change
    JFreeChartUtil.fireChangeEvent(renderer);
    
  }
  
  
  /**
   * Applies style properties which can only be applied to
   * {@link XYBarRenderer}.<br>
   * Called by {@link #applyToRenderer(AbstractRenderer)}. 
   * @param renderer a {@link XYBarRenderer}
   */
  protected void applyToBarRenderer(XYBarRenderer renderer) {
    if ( getMargin() != null )
      renderer.setMargin(margin);
  }

  /**
   * Applies style properties which can only be applied to
   * {@link AbstractXYItemRenderer}.<br>
   * Called by {@link #applyToRenderer(AbstractRenderer)}. 
   * @param renderer a {@link AbstractXYItemRenderer}
   */
  protected void applyToXYItemRenderer(AbstractXYItemRenderer renderer) {
    renderer.setLegendItemLabelGenerator( getLegendLabelGenerator() );
    renderer.setLegendItemToolTipGenerator( getLegendTooltipGenerator() );
  }

  /**
   * Applies style properties which can only be applied to
   * {@link AbstractCategoryItemRenderer}.<br>
   * Called by {@link #applyToRenderer(AbstractRenderer)}. 
   * @param renderer a {@link AbstractCategoryItemRenderer}
   */
  protected void applyToCategoryItemRenderer(AbstractCategoryItemRenderer renderer) {
    renderer.setLegendItemLabelGenerator( getLegendLabelGenerator() );
    renderer.setLegendItemToolTipGenerator( getLegendTooltipGenerator() );
  }
  
  
// TODO: Fuer PieCharts hier weiter implementieren
//  /**
//   * Applies the rendering style to a pie plot.<br>
//   * <b>Note</b>:
//   * <ul>
//   *   <li>Because PieCharts have no multiple series, only the "series" style
//   *       with index 0 (of the RendererStyle) is taken into account./li>
//   * </ul>
//   * @param renderer a renderer
//   */
//  public void applyToRenderer(PiePlot piePlot) {
//    if (piePlot == null) {
//      LOGGER.debug("No renderer to be customised.");
//      return;
//    }
//
//    // Apply item label visibility (only the first "series" is
//    // taken into account)
//    // TODO: can not be applied to PiePlot, only the lines can be
//    //       set in/visible
//    if ( isSeriesItemLabelsVisible(0) != null )
//      piePlot.setLabelLinksVisible(isSeriesItemLabelsVisible(0));
//    
//    // Apply shape visibility
//    for (int series : seriesShapesVisible.keySet())
//      try {
//        if ( isSeriesShapesVisible(series) != null ) {
//            final Shape shape = isSeriesShapesVisible(series) ? 
//              AbstractRenderer.DEFAULT_SHAPE
//              : null;
//            renderer.setSeriesShape(series,
//                      shape,
//                      false);
////          LOGGER.debug("Setting shape = "+shape +" for series "+series);
//        }
//        
//      } catch (Exception err) {
//        LOGGER.warn("Renderer contains not series "+series+". Shapes visibility can not be applied.");
//      }
//
//    // Apply rendering color
//    for (int series : seriesPaint.keySet())
//      try {
//        if ( getSeriesPaint(series) != null )
//          renderer.setSeriesPaint(series, getSeriesPaint(series), false);
//      } catch (Exception err) {
//        LOGGER.warn("Renderer contains not series "+series+". Paint can not be applied.");
//      }
//
//    // Apply legend visibility
//    for (int series : seriesLegendVisible.keySet())
//      try {
//        if ( isSeriesLegendVisible(series) != null )
//          renderer.setSeriesVisibleInLegend(series, isSeriesLegendVisible(series), false);
//      } catch (Exception err) {
//        LOGGER.warn("Renderer contains not series "+series+". Legend visibility can not be applied.");
//      }
//
//    // Other properties must be applied according to the specific renderer type
//    if (renderer instanceof XYBarRenderer)
//      applyToBarRenderer((XYBarRenderer)renderer);
//    if (renderer instanceof AbstractXYItemRenderer)
//      applyToXYItemRenderer((AbstractXYItemRenderer)renderer);
//    if (renderer instanceof AbstractCategoryItemRenderer)
//      applyToCategoryItemRenderer((AbstractCategoryItemRenderer)renderer);
//    
//    // notify the listeners about style change
//    JFreeChartUtil.fireChangeEvent(renderer);
//    
//  }

  /**
   * This class implements a label generator based on the
   * <b>legend label</b> definitions of this style, which can be used also
   * for {@link AbstractCategoryItemRenderer} and {@link AbstractXYItemRenderer}.
   * If no label definition is defined in the style, the label generator
   * falls back to the standard label generators.
   * @see ChartRendererStyle#LABELGEN_CAT
   * @see ChartRendererStyle#LABELGEN_XY
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
   */
  protected class LegendLabelGenerator implements XYSeriesLabelGenerator, CategorySeriesLabelGenerator {
    /**
     * Returns the legend label for a series.
     * @param dataset only used for the standard label generator
     * @param series  the series the legend label is returned for
     * @see ChartRendererStyle#getSeriesLegendLabel(int)
     */
    @Override
    public String generateLabel(XYDataset dataset, int series) {
      if ( getSeriesLegendLabel(series) != null )
        return getSeriesLegendLabel(series).getLabel();
      if ( getDefaultLegendLabel() != null )
        return getDefaultLegendLabel().getLabel();
      // if label is not defined, use the standard label generator 
      return LABELGEN_XY.generateLabel(dataset, series);
    }

    /**
     * Returns the legend label for a series.
     * @param dataset only used for the standard label generator
     * @param series  the series the legend label is returned for
     * @see ChartRendererStyle#getSeriesLegendLabel(int)
     */
    @Override
    public String generateLabel(CategoryDataset dataset, int series) {
      if ( getSeriesLegendLabel(series) != null )
        return getSeriesLegendLabel(series).getLabel();
      if ( getDefaultLegendLabel() != null )
        return getDefaultLegendLabel().getLabel();
      // if label is not defined, use the standard label generator 
      return LABELGEN_CAT.generateLabel(dataset, series);
    }
  }

  /**
   * This class implements a label generator based on the
   * <b>legend tooltip</b> definitions of this style, which can be used also
   * for {@link AbstractCategoryItemRenderer} and {@link AbstractXYItemRenderer}.
   * If no tooltip definition is defined in the style, the label generator
   * generates <b>no tooltip</b>.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
   */
  protected class LegendTooltipGenerator implements XYSeriesLabelGenerator, CategorySeriesLabelGenerator {
    /**
     * Returns the legend tooltip for a series.
     * @param dataset not used
     * @param series  the series the legend label is returned for
     * @return {@code null} if the style does not define a tooltip
     *         for the series
     * @see ChartRendererStyle#getSeriesLegendTooltip(int)
     */
    @Override
    public String generateLabel(XYDataset dataset, int series) {
      return generateLabel(series);
    }

    /**
     * Returns the legend label for a series.
     * @param dataset not used
     * @param series  the series the legend label is returned for
     * @return {@code null} if the style does not define a tooltip
     *         for the series
     * @see ChartRendererStyle#getSeriesLegendTooltip(int)
     */
    @Override
    public String generateLabel(CategoryDataset dataset, int series) {
      return generateLabel(series);
    }
    
    public String generateLabel(int series) {
      if ( getSeriesLegendTooltip(series) != null )
        return getSeriesLegendTooltip(series).getLabel();
      if ( getDefaultLegendTooltip() != null )
        return getDefaultLegendTooltip().getLabel();
      // if label is not defined, generate no tooltip 
      return null;
    }
  }
  
}
