/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/
/*
 *  SDSS Framework - This file is part of the Spatial Decision Support System Framework Platform
 *
 *  BasicChartStyle.java
 *
 *  Responsible developper: Andreas Enders, Geographical Institut, Hydrology Research Group, Meckenheimer Allee 166, 53115 Bonn, Germany
 *                                          IMPETUS Project, www.impetus.uni-koeln.de
 *  Contact Information:                    info@andreas-enders.de
 */

package de.schmitzm.jfree.chart.style;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.general.PieDataset;
import org.jfree.data.xy.XYDataset;

import de.schmitzm.lang.LangUtil;

/**
 * This class provides a basis chart style for pie charts.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class PieChartStyle extends AbstractChartStyle {

  /**
   * Creates a normal pie style with default values.
   * @param id   a (unique) ID for the style
   */
  public PieChartStyle(String id) {
    this(id, null, null, null, true, true, true);
  }

  /**
   * Creates a chart style.
   * @param id       a (unique) ID for the style
   * @param title    title for the chart
   * @param xTitle   title for the chart's X-axis
   * @param yTitle   title for the chart's Y-axis
   * @param legend   flag whether a legend is generated
   * @param tooltips flag whether toolstips are generated
   * @param urls     flag whether URLs are generated
   */
  public PieChartStyle(String id, 
                       String title,   String xTitle,    String yTitle,
                       boolean legend, boolean tooltips, boolean urls) {
    super(id, ChartType.PIE, title, xTitle, yTitle, legend, tooltips, urls);
  }

  /**
   * Creates a (deep) clone of this chart style.
   */
  @Override
  public PieChartStyle copy() {
    return (PieChartStyle) copyTo( new PieChartStyle("") );
  }
  
  /**
   * Copies all properties of this chart style to another
   * chart style.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   * @exception IllegalArgumentException if {@code dest} is no
   *            {@link AbstractChartStyle}
   */
  @Override
  public ChartStyle copyTo(ChartStyle dest) {
    // copy the super class properties
    dest = super.copyTo(dest);
    if ( dest instanceof PieChartStyle ) {
      PieChartStyle destBCS = (PieChartStyle) dest;
      // copy this classes properties
      // ...
    }
    return dest;
  }

  /**
   * Creates a chart for the given {@link Dataset} and applies the style.
   * @param dataset the data for the chart
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the given dataset
   */
  public JFreeChart applyToDataset(Dataset... dataset) {
    if ( dataset.length > 1 )
      LOGGER.warn("PieChart only supports one dataset. Secondary dataset will be ignored!");
    if ( !(dataset[0] instanceof PieDataset) )
      throw new UnsupportedOperationException(LangUtil.getSimpleClassName(this)+" can only be applied to PieDataset!");
    return applyToDataset((PieDataset)dataset[0]);
  }

  /**
   * Creates a chart for the given {@link Dataset} and applies the style.
   * @param dataset the data for the chart
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the given dataset
   */
  public JFreeChart applyToDataset(PieDataset dataset) {
    JFreeChart chart = ChartFactory.createPieChart(null, dataset, legend, tooltips, urls);
    applyToChart(chart);
    return chart;  }
  
  /**
   * Applies the style to an existing chart.
   * @param chart chart the style is applied to
   */
  public void applyToChart(JFreeChart chart) {
    // apply the pie chart properties
    // ...
    super.applyToChart(chart);
  }
  

  /**
   * Creates a line chart from an {@link XYDataset} or {@link CategoryDataset}
   * @param dataset Dataset
   */
  private JFreeChart createLineChart(Dataset dataset) {
    JFreeChart chart = null;
    // XYDataset -> XYLineChart
    if (dataset instanceof XYDataset) {
      chart = ChartFactory.createXYLineChart(
          null,null,null, // Title and axis will be configured in applyToChart(..)
          (XYDataset)dataset,
          orientation,
          legend,
          tooltips,
          urls
      );
    }
    // CategoryDataset -> LineChart
    if (dataset instanceof CategoryDataset) {
      chart = ChartFactory.createLineChart(
          null,null,null, // Title and axis will be configured in applyToChart(..)
          (CategoryDataset)dataset,
          orientation,
          legend,
          tooltips,
          urls
      );
    }
    return chart;
  }
}
