/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package de.schmitzm.jfree.data;

import org.jfree.data.DomainOrder;
import org.jfree.data.general.DatasetChangeListener;
import org.jfree.data.general.DatasetGroup;
import org.jfree.data.xy.XYDataset;

/**
 * This class simply pipes the calls of an {@link XYDataset}.
 * This class can be used to extend an existing {@link XYDataset}
 * with new functionalities.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class PipedXYDataset implements XYDataset {
  /** Holds the original dataset */
  protected XYDataset dataset = null;
  
  /**
   * Creates a new dataset.
   * @param dataset the original data.
   */
  public PipedXYDataset(XYDataset dataset) {
    this.dataset = dataset;
  }
  
  /**
   * Returns the original data which is piped by this
   * class.
   */
  public XYDataset getBaseDataset() {
    return dataset;
  }
  
  /**
   * Returns the series count. 
   */
  @Override
  public int getSeriesCount() {
    return dataset.getSeriesCount();
  }

  /**
   * Returns the series key of the i-th series.
   */
  @Override
  public Comparable getSeriesKey(int series) {
    return dataset.getSeriesKey(series);
  }

  /**
   * Returns the series index for a given series key.
   */
  @Override
  public int indexOf(Comparable seriesKey) {
    return dataset.indexOf(seriesKey);
  }

  /**
   * Adds a change listener to the dataset.
   */
  @Override
  public void addChangeListener(DatasetChangeListener listener) {
    dataset.addChangeListener(listener);
  }

  /**
   * Removes a change listener from the dataset.
   */
  @Override
  public void removeChangeListener(DatasetChangeListener listener) {
    dataset.removeChangeListener(listener);
  }
  
  /**
   * Returns the dataset group of the dataset.
   */
  @Override
  public DatasetGroup getGroup() {
    return dataset.getGroup();
  }


  /**
   * Sets the dataset group of the dataset.
   */
  @Override
  public void setGroup(DatasetGroup group) {
    dataset.setGroup(group);
  }

  /**
   * Returns the order of the domain (or X) values returned by the dataset. 
   */
  @Override
  public DomainOrder getDomainOrder() {
    return dataset.getDomainOrder();
  }

  /**
   * Returns the number of items in a series.
   */
  @Override
  public int getItemCount(int series) {
    return dataset.getItemCount(series);
  }

  /**
   * Returns the x-value for an item within a series.
   * @param series the series index
   * @param item the item index within the series 
   */
  @Override
  public Number getX(int series, int item) {
    return dataset.getX(series, item);
  }

  /**
   * Returns the x-value for an item within a series.
   * @param series the series index
   * @param item the item index within the series 
   */
  @Override
  public double getXValue(int series, int item) {
    return dataset.getXValue(series, item);
  }

  /**
   * Returns the y-value for an item within a series.
   * @param series the series index
   * @param item the item index within the series 
   */
  @Override
  public Number getY(int series, int item) {
    return dataset.getY(series, item);
  }

  /**
   * Returns the y-value for an item within a series.
   * @param series the series index
   * @param item the item index within the series 
   */
  @Override
  public double getYValue(int series, int item) {
    return dataset.getYValue(series, item);
  }

}
