/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package de.schmitzm.jfree.data;

import java.util.Arrays;

import org.jfree.data.function.Function2D;
import org.jfree.data.function.LineFunction2D;
import org.jfree.data.general.DatasetUtilities;
import org.jfree.data.xy.XYDataset;

import de.schmitzm.jfree.JFreeChartUtil;

/**
 * An implementation of {@link RegressionDataset}. Because of technical restrictions
 * this class has only internal an constructor, so you have to create instances
 * by calling {@link #create(XYDataset, int, String, double, double, int)}. 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class RegressionDatasetImpl extends PipedXYDataset implements RegressionDataset {
  /** Holds the number of data items the regression calculation
   *  bases on. */
  protected int      dataCount = 0;
  /** Holds the R value of the regression. */
  protected double   rValue = 0.0;
  /** Holds the coefficients of the regression line (0 = intercept, 1 = slope). */
  protected double[] olsCoefficients = null;
  /** Holds the regression line. */
  protected Function2D regrLine = null;
  
  /**
   * Constructs a new instance. Only the regression dateset
   * is set by this constructor! The additional informations (R
   * and coefficients) remains empty!!
   * @param dataset
   */
  protected RegressionDatasetImpl(XYDataset dataset) {
    super(dataset);
  }
  
  /**
   * Creates a new {@link RegressionDataset}
   * @param dataset     the dataset which hold the data the regression line is calculated from
   * @param series      the series in the dataset the regression line is calculated for
   * @param seriesKey   the series key for the created dataset
   * @param startX      the first value the sample data is created for
   * @param endX        the last value the sample data is created for
   * @param sampleCount count of created samples
   */
  public static RegressionDataset create(XYDataset dataset,
      int series, String seriesKey, double startX, double endX,
      int sampleCount) {
    
    // (pre)calculate the parameters 
    double[] coefficients = JFreeChartUtil.getOLSRegression(dataset, series);
    Function2D curve      = new LineFunction2D(coefficients[0], coefficients[1]);
    double     r          = coefficients[2];
    int        dataCount  = (int)coefficients[3];
    
    // Put additional legend informations in series key
    // NOTE: The KECK projects parses this information!!
    //       So be aware of uncontrolled changes to this
    //       information (or its format)!!!
    seriesKey += ", r=" + JFreeChartUtil.getRFormatted(r);
    seriesKey += " (" + JFreeChartUtil.getRQuality(r) + ")";

    XYDataset  regrData   = DatasetUtilities.sampleFunction2D(
        curve, startX, endX, sampleCount,seriesKey);
    
    // create and initialize the regression dataset
    RegressionDatasetImpl regrDataset = new RegressionDatasetImpl(regrData);
    regrDataset.regrLine = curve;
    regrDataset.olsCoefficients = new double[] {coefficients[0], coefficients[1]};
    regrDataset.rValue = r;
    regrDataset.dataCount = dataCount;
        
    return regrDataset;
  }
  
  /**
   * Returns the regression line for the dataset.
   */
  public Function2D getRegressionLine() {
    return regrLine;
  }
  
  /**
   * Returns the R-Value of the regression line, which is the
   * product-moment-correlation coefficient and gives information
   * about how strong the variables are related.
   */
  @Override
  public double getR() {
    return rValue;
  }

  /**
   * Returns the number of samples the regression calculation bases
   * on.
   */
  public int getSampleCount() {
    return dataCount;
  }

  /**
   * Returns the 2 regression coefficients: [0] = the intercept; [1] = the slope
   * <b>Note</b>: Each call returns a new array instance!
   */
  @Override
  public double[] getOLSRegressionCoefficients() {
    return Arrays.copyOf(olsCoefficients,olsCoefficients.length);
  }

  /**
   * Returns the slope of the regression line.
   */
  public double getRegressionLineSlope() {
    return olsCoefficients[1];
  }

  /**
   * Returns the intercept of the regression line.
   */
  public double getRegressionLineIntercept() {
    return olsCoefficients[0];
  }

  /**
   * Returns a formatted representation of the R-value.
   * @see #getR()
   * @see JFreeChartUtil#getRFormatted(double)
   */
  public String getRFormatted() {
    return JFreeChartUtil.getRFormatted( getR() );
  }

  /**
   * Returns a quality representation of the R-value (e.g. weak,
   * medium, strong, ...).
   * @see #getR()
   * @see JFreeChartUtil#getRQuality(double)
   */
  public String getRQuality() {
    return JFreeChartUtil.getRQuality( getR() );
  }

}
