package de.schmitzm.jfree.table;

import de.schmitzm.jfree.JFreeChartUtil;
import de.schmitzm.jfree.table.style.TableChartStyle;
import de.schmitzm.lang.ResourceProvider;

/**
 * Defines possible aggregation function to calculate on attribute categories.
 * @see TableChartStyle#setAttributeAggregation(int, AggregationFunction)
 * @see TableChartStyle#getAttributeAggregation(int)
 * @see optionally TableChartStyle#setNoDataWeightValues(int, java.util.Set))
 * @see optionally
 *      TableChartStyle#getAttributeAggregationWeightAttributeName(int)
 */
public enum AggregationFunction {

  /** Count of the attribute values. */
  COUNT(),
  /** Sum of the attribute values. */
  SUM(),
  /** Sum of the absolute attribute values. */
  SUM_ABS(true,false, false),
  /** Weighted sum of the attribute values. Needs an attribute for weighting. */
  SUM_WEIGHTED(false, true, false),
  /** Average of the attribute values. */
  AVG(),
  /**
   * Weighted average of the attribute values. Needs an attribute for weighting.
   */
  AVG_WEIGHTED(false, true, true),
  /** Median of the attribute values. */
  MEDIAN(),
  /** Minimum attribute value. */
  MIN(),
  /** Maximum attribute value. */
  MAX(),
  /** Variance of the attribute values. */
  VARIANCE(),
  /** Standard deviation of the attribute values. */
  STND_DEV();

  // Indicates whether the attribute is must be transformed to positive
  // values
  private final boolean absoluted;
  // Indicates whether the attribute is weighted with another attribute
  private final boolean weighted;
  // Indicates whether the attribute is weighted and averaged with the weight
  // sum
  private final boolean averaged;

  /**
   * Creates a default {@link Enum}, which is not absoluted, not weighted and
   * not averaged.
   */
  private AggregationFunction() {
    this(false,false,false);
  }

  /**
   * Creates an {@link Enum} instance.
   * @param absoluted Indicates whether the attribute is must be transformed to
   *          positive values
   * @param weighted Indicates whether the attribute is weighted with another
   *          attribute
   * @param averaged Indicates whether the attribute is weighted and averaged
   *          with the weight sum (if averaged, weighted is automatically
   *          implied!)
   */
  private AggregationFunction(boolean absoluted, boolean weighted,
      boolean averaged) {
    this.absoluted = absoluted;
    this.weighted = weighted || averaged; // average implies weighted
    this.averaged = averaged;
  }

  /**
   * Returns <code>true</code>, if this aggregation method needs the
   * absolute attribute value.
   * @see Math#abs(double)
   **/
  public boolean isAbsoluted() {
    return absoluted;
  }

  /**
   * Returns <code>true</code>, if this aggregation method is used a second
   * attribute for weighting.
   **/
  public boolean isWeighted() {
    return weighted;
  }

  /**
   * Returns <code>true</code>, if this aggregation method is used a second
   * attribute for averaging.
   **/
  public boolean isAveraged() {
    return averaged;
  }

  /**
   * Prefix for the title (.TITLE) and description (.DESC) key in the resource
   * bundle.
   * @see #getTitle()
   * @see #getDescription()
   */
  public final String RESOURCE_PREFIX = getClass().getSimpleName() + "." +
                                        toString();

  /**
   * Returns a description of this kind of chart. Can be used for tool-tips. May
   * return <code>null</code> if no localized String found.
   */
  public String getDescription() {
    final String resource = JFreeChartUtil.R(RESOURCE_PREFIX + ".Desc");
    if (resource == null ||
        resource.equals(ResourceProvider.MISSING_RESOURCE_STRING))
      return null;
    return resource;
  }

  /**
   * Returns a localized title of this kind of chart. If no localized string
   * found, return the Enum.toString()
   */
  public String getTitle() {
    final String resource = JFreeChartUtil.R(RESOURCE_PREFIX + ".Title");
    if (resource == null ||
        resource.equals(ResourceProvider.MISSING_RESOURCE_STRING))
      return toString();
    return resource;
  }

}
