/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/

package de.schmitzm.jfree.table.style;

import java.awt.Color;

import org.jfree.chart.axis.Axis;

import de.schmitzm.i18n.Translation;
import de.schmitzm.jfree.JFreeChartUtil;
import de.schmitzm.jfree.chart.style.ChartAxisStyle;
import de.schmitzm.lang.LangUtil;

/**
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *
 */
public class TableChartAxisStyle extends ChartAxisStyle {
  /** Holds the {@link TableChartStyle} the axis is connected to. */
  protected TableChartStyle<?> chartStyle = null;
  
  /**
   * Creates a new style with default values (empty label, color black, angle 0).
   * @param chartStyle the chart style the axis is connected to
   */
  public TableChartAxisStyle(TableChartStyle<?> chartStyle) {
    this(chartStyle,"",Color.black,0.0,0.0);
  }

  /**
   * Creates a new style.
   * @param chartStyle the chart style the axis is connected to
   * @param title axis title
   * @param color text color for the axis title
   * @param labelAngle angel (in degrees) the axis label is rotated by
   * @param valuesAngle angel (in degrees) the axis values are rotated by
   */
  public TableChartAxisStyle(TableChartStyle<?> chartStyle, String title, Color color, Double labelAngle, Double valuesAngle) {
    this(chartStyle, new Translation(title), color, labelAngle, valuesAngle);
  }

  /**
   * Creates a new style.
   * @param chartStyle the chart style the axis is connected to
   * @param title axis title as an internationalized {@link Translation} 
   * @param color text color for the axis title
   * @param labelAngle angel (in degrees) the axis label is rotated by
   * @param valuesAngle angel (in degrees) the axis values are rotated by
   */
  public TableChartAxisStyle(TableChartStyle<?> chartStyle, Translation title, Color color, Double labelAngle, Double valuesAngle) {
    super(title, color, labelAngle, valuesAngle);
    this.chartStyle = chartStyle; 
  }
  
  /**
   * Creates a (deep) clone of this style.
   */
  @Override
  public TableChartAxisStyle copy() {
    return (TableChartAxisStyle)copyTo( new TableChartAxisStyle(null) );
  }

  /**
   * Creates a (deep) clone of this style.
   */
  @Override
  public ChartAxisStyle copyTo(ChartAxisStyle dest) {
    dest = super.copyTo(dest);
    // also copy the chart style, the axis is connected to
    if ( dest instanceof TableChartAxisStyle ) {
      ((TableChartAxisStyle)dest).setChartStyle( getChartStyle() );
    }
    return dest;
  };
  
  /**
   * Returns the chart style, the axis is connected to.
   */
  public TableChartStyle<?> getChartStyle() {
    return chartStyle;
  }

  /**
   * Sets the chart style, the axis is connected to.
   */
  public void setChartStyle(TableChartStyle<?> style) {
    chartStyle = style;
  }
  
  /**
   * Applies the style to a specific axis.<br>
   * Note: Some style properties can only be applied to specific axis types.
   * @param axis an axis
   */
  @Override
  public void applyToAxis(Axis axis) {
    super.applyToAxis(axis);
    if (axis == null) {
      return;
    }

    // Determine the unit, aggregation function title and normalization
    // to display with the axis title
    String unitStr = isUnitVisible() ? getUnitString() : null;
    String funcStr = null;
    String normStr = null; 
    for (int i=0; chartStyle != null && i<chartStyle.getAxisCount(); i++)
      if ( chartStyle.getAxisStyle(i) == this ) {
        // TODO: ??? von welchem Attribut die Aggr.Function und
        //       Normalisierung verwenden. Ist 'i' richtig?
        if ( chartStyle.getAttributeAggregation(i) != null )
          funcStr = chartStyle.getAttributeAggregation(i).getTitle();
        if ( chartStyle.isAttributeNormalized(i) )
          normStr = JFreeChartUtil.R("FeatureChartAxisStyle.normalized");
        break;
      }
    
    // If unit, function or normalization is set, add it to axis title
    String axisLabel = getLabel();
    String unitLabel = LangUtil.listString(", ",unitStr,funcStr,normStr);
    if ( unitLabel != null && !"".equals(unitLabel) )
      axisLabel += " ["+unitLabel.trim()+"]";
    axis.setLabel( axisLabel );
  }

}
