/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/
package de.schmitzm.jfree.table.style;

import java.util.List;
import java.util.Set;

import org.jdom.Element;

import de.schmitzm.jfree.chart.style.ChartAxisStyle;
import de.schmitzm.jfree.chart.style.ChartStyle;
import de.schmitzm.jfree.chart.style.ChartStyleXMLFactory;
import de.schmitzm.jfree.chart.style.ChartType;
import de.schmitzm.jfree.table.AggregationFunction;
import de.schmitzm.xml.XMLUtil;
	
/**
 * This class defines a factory to create a chart style from XML
 * specialized for table data.
 * @see TableChartStyle
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public abstract class TableChartStyleXMLFactory<E extends TableChartStyle<?>> extends ChartStyleXMLFactory<E> {

  /**
   * Applies the chart style definition from XML (<b>except the type and id!!</b>)
   * to an existing {@link ChartStyle} object.
   * @param chartStyle an existing chart style ({@code null} not permitted!)
   * @param element    element to read the properties from
   */
  @Override
  public void applyStyleFromXML(E chartStyle, Element element) {
    // Apply the normal style properties
    super.applyStyleFromXML(chartStyle, element);
    
    // Convert ChartAxisStyle to TableChartAxisStyle
    for (int i=0; i<chartStyle.getAxisCount(); i++) {
      ChartAxisStyle axisStyle = chartStyle.getAxisStyle(i);
      if ( axisStyle != null ) {
        TableChartAxisStyle tableAxisStyle = (TableChartAxisStyle)axisStyle.copyTo( new TableChartAxisStyle(chartStyle) );
        chartStyle.setAxisStyle(i, tableAxisStyle);
      }
    }
    
    // Attributes of <tableAttributes>
    XMLUtil.getBooleanAttribute(element, "ignoreNullRangeValues", true);
    
    // Attribute definitions in <tableAttributes> element
    Element attributesElement = element.getChild("tableAttributes");
    if ( attributesElement == null )
    	throw new UnsupportedOperationException("<tableAttributes> element necessary for TableChartStyle!");
    
//      // try old-fashioned XML-definition with <featureAttributes>
//      // DONE: remove this temporary work-around!
//      attributesElement = element.getChild("featureAttributes");
//    if ( attributesElement == null )

    // Read domain attribute
    Element domainAttrElem = attributesElement.getChild("domainAttr");
    if ( domainAttrElem == null )
      throw new UnsupportedOperationException("Attribute <domainAttr> required for TableChartStyle");
    String domainAttrName = XMLUtil.getAttribute(domainAttrElem, "name");
    if ( domainAttrName == null )
      throw new UnsupportedOperationException("Attribute 'name' necessary for <domainAttr> element of TableChartStyle!");
    chartStyle.setAttributeName(0, domainAttrName);
    chartStyle.setAttributeNormalized(0, XMLUtil.getBooleanAttribute(domainAttrElem, "normalize", false));
    chartStyle.setSortDomainAxis(XMLUtil.getBooleanAttribute(domainAttrElem, "sort", false));
    chartStyle.setForceCategories(XMLUtil.getBooleanAttribute(domainAttrElem, "forceCategories", false));
    Set<Object> nullAliases = XMLUtil.getSetAttribute(domainAttrElem, "nullAliases");
    if (nullAliases != null)
      chartStyle.setNoDataValues(0, nullAliases);
    
    // Read all range attribute definitions
    int rangeAttrNo = 0;
    for (Element rangeAttrElem : (List<Element>)attributesElement.getChildren("rangeAttr") ) {
      String tableAttrName = XMLUtil.getAttribute(rangeAttrElem, "name");
      if ( tableAttrName == null )
        throw new UnsupportedOperationException("Attribute 'name' necessary for <rangeAttr> element of TableChartStyle!");
      // Apply attributes to style
      rangeAttrNo++;
      chartStyle.setAttributeName(rangeAttrNo, tableAttrName);
      chartStyle.setAttributeNormalized(rangeAttrNo, XMLUtil.getBooleanAttribute(rangeAttrElem, "normalize", false));
      nullAliases = XMLUtil.getSetAttribute(rangeAttrElem, "nullAliases");
      if (nullAliases != null)
    	  chartStyle.setNoDataValues(rangeAttrNo, nullAliases);
      
      String aggrFuncStr = XMLUtil.getAttribute(rangeAttrElem, "function", (String)null);
      if ( aggrFuncStr != null ) 
        chartStyle.setAttributeAggregation(rangeAttrNo, AggregationFunction.valueOf(aggrFuncStr) );

      String aggrFuncWeightAtt = XMLUtil.getAttribute(rangeAttrElem, "weightAttrName", (String)null);
      if ( aggrFuncWeightAtt != null )
        chartStyle.setAttributeAggregationWeightAttributeName(rangeAttrNo, aggrFuncWeightAtt );
      
      nullAliases = XMLUtil.getSetAttribute(rangeAttrElem, "weightAttrNullAliases");
      if (nullAliases != null)
    	  chartStyle.setWeightAttributeNoDataValues(rangeAttrNo, nullAliases);
    }

    // Read (optional) series attribute (its values defines the series groups)
    Element seriesAttrElem = attributesElement.getChild("seriesAttr");
    if ( seriesAttrElem != null ) {
      String seriesAttrName = XMLUtil.getAttribute(seriesAttrElem, "name");
      if ( seriesAttrName == null )
        throw new UnsupportedOperationException("Attribute 'name' necessary for <seriesAttr> element of TableChartStyle!");
      chartStyle.setSeriesAttributeName(seriesAttrName);
      nullAliases = XMLUtil.getSetAttribute(seriesAttrElem, "nullAliases");
      if (nullAliases != null)
        chartStyle.setSeriesAttributeNoDataValues(nullAliases);
      String seriesLegendLableAttrName = XMLUtil.getAttribute(seriesAttrElem, "titleAttrName");
      if (seriesLegendLableAttrName != null)
        chartStyle.setSeriesLegendTitleAttributeName(seriesLegendLableAttrName);
    }
  }


  /**
   * Creates a default style for a chart type.
   * @param id   a (unique) ID for the style
   * @param type a chart type
   */
  @Override
  public E createDefaultChartStyle(String id, ChartType type) {
    throw new UnsupportedOperationException("TableChartStyleXMLFactory can not create default styles, because the TableChartStyle implentation ist abstract for type "+type);
  }

  /**
   * Creates a XML {@link Element} representing the {@link TableChartStylee}.
   * @param style           style to create an element for
   * @param rootElementName name for the root element (if {@code null} "ChartType" is 
   *                        used)
   */
  @Override
  public Element createElementFromStyle(E style, String rootElementName) {
    Element root = super.createElementFromStyle(style, rootElementName);
    
    // Create and add element <tableAttributes>
    Element attrElem = addChildToElement(root, "tableAttributes");
    
    // Ignore NoData range values or treat as numeric 0.
    XMLUtil.setAttribute(attrElem, "ignoreNullRangeValues", style.isIgnoreNoDataRangeValues(),false);
    
    // Domain attribute definition
    if ( style.getAttributeCount() > 0 )
      addChildToElement(attrElem, "domainAttr", false,
          "name", style.getAttributeName(0),
          "normalize", style.isAttributeNormalized(0),
          "nullAliases", XMLUtil.convertSetToSeparatedString(style.getNoDataValues(0)), 
          "sort", style.isSortDomainAxis(),
          "forceCategories", style.isForceCategories()
      );
    
    // Range attributes definition
    for (int i=1; i<style.getAttributeCount(); i++)
      addChildToElement(attrElem, "rangeAttr", false,
          "name", style.getAttributeName(i),
          "normalize", style.isAttributeNormalized(i),
          "nullAliases", XMLUtil.convertSetToSeparatedString(style.getNoDataValues(i)), 
          "function", style.getAttributeAggregation(i),
          "weightAttrName", style.getAttributeAggregationWeightAttributeName(i),
          "weightAttrNullAliases", XMLUtil.convertSetToSeparatedString(style.getWeightAttributeNoDataValues(i))
      );

    // Series attribute definition
    if ( style.getSeriesAttributeName() != null )
      addChildToElement(attrElem, "seriesAttr", false,
          "name", style.getSeriesAttributeName(),
          "nullAliases", XMLUtil.convertSetToSeparatedString(style.getSeriesAttributeNoDataValues()),
          "titleAttrName", style.getSeriesLegendTitleAttributeName()
      );

    return root;
  }
}
