/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Tzeggai - additional utility classes
 ******************************************************************************/
package de.schmitzm.jfree.table.style;

import java.util.Set;

import org.jfree.chart.JFreeChart;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;

import de.schmitzm.jfree.chart.style.ChartStyle;
import de.schmitzm.jfree.chart.style.ScatterChartStyle;
import de.schmitzm.jfree.table.AggregationFunction;

/**
 * This class extends the {@link ScatterChartStyle} with the properties
 * of the {@link TableChartStyle}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public abstract class TableScatterChartStyle<E> extends ScatterChartStyle implements TableChartStyle<E> {
  /** Used to maintain the {@link TableChartStyle} properties. */
  protected Dummy<E> dummyTableChartStyle = null;
  
  /**
   * Creates a scatter chart style a regression line shown.
   * @param id   a (unique) ID for the style
   */
  public TableScatterChartStyle(String id) {
    super(id);
    dummyTableChartStyle = new Dummy(id,2);
  }
  
  @Override
  /**
   * Nicer debugging
   */
	public String toString() {
	  
	  String titel = getTitleStyle() != null ? getTitleStyle().getLabel() : " - null title!!! - ";
		return "ID="+getID()+"   "+titel+"" + super.toString();
	}

  /**
   * Creates a scatter chart style a regression line shown.
   * @param id         a (unique) ID for the style
   * @param title      title for the chart
   * @param xAttrName  table attribute used for domain axis
   * @param yAttrName  table attribute used for range axis
   */
  public TableScatterChartStyle(String id, String title, String xAttrName, String yAttrName) {
    this(id);
    setAttributeName(0, xAttrName);
    setAttributeName(1, yAttrName);
    getTitleStyle().setLabel(title);
  }

  /**
   * Copies all properties of this style to another one.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   */
  @Override
  public ChartStyle copyTo(ChartStyle dest) {
    // copy the super class properties
    dest = super.copyTo(dest);
    // copy this classes properties
    if ( dest instanceof TableChartStyle )
      dummyTableChartStyle.copyTo(dest);

    return dest;
  }

  /**
   * Removes all style informations about an attribute and
   * reorganizes the attribute indexes so there is an continuous 
   * order. 
   * @param idx an attribute
   */
  @Override
  public void removeAttribute(int idx) {
    dummyTableChartStyle.removeAttribute(idx);
  }

  /**
   * Returns the maximum number of table attributes that can be
   * specified by this style.
   * @return always 2
   */
  @Override
  public int getMaxAttributeCount() {
    return dummyTableChartStyle.getMaxAttributeCount();
  }

  /**
   * Returns the number of table attributes defined in this style.
   */
  @Override
  public int getAttributeCount() {
    return dummyTableChartStyle.getAttributeCount();
  }

  /**
   * Returns the name of a table attribute needed to create a
   * chart for this style.
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   */
  @Override
  public String getAttributeName(int idx) {
    return dummyTableChartStyle.getAttributeName(idx);
  }

  /**
   * Sets the name of a table attribute needed to create a
   * chart for this style.
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @param attrName table attribute name 
   */
  @Override
  public void setAttributeName(int idx, String attrName) {
    dummyTableChartStyle.setAttributeName(idx,attrName);
  }

  /**
   * Returns the axis an attribute deals with. 
   * @param idx attribute index (0=domain; 1=1st range series; 2=2nd range
   *          series; ...)
   * @return always {@link ChartStyle#DOMAIN_AXIS} for domain axis; {@link ChartStyle#RANGE_AXIS}
   *         for all other attribute for which the axis is not set explicitly 
   */
  @Override
  public int getAttributeAxis(int idx) {
    return dummyTableChartStyle.getAttributeAxis(idx);
  }
  
  /**
   * Sets the axis an attribute deals with.<br>
   * <b>Note:</b> Does nothing for the domain attribute (0), because
   * the domain attribute always deals with the domain axis automatically! 
   * @param idx attribute index (0=domain; 1=1st range series; 2=2nd range
   *          series; ...)
   * @param axis axis number ({@link ChartStyle#RANGE_AXIS}, {@link ChartStyle#RANGE_AXIS2})
   * @exception IllegalArgumentException if an invalid axis is specified
   */
  @Override
  public void setAttributeAxis(int idx, Integer axis) {
    dummyTableChartStyle.setAttributeAxis(idx, axis);
  }

  /**
   * Checks how many range axis are defined by the attributes.
   */
  @Override
  public int determineRangeAxisCount() {
    return dummyTableChartStyle.determineRangeAxisCount();
  }

  /**
   * Sets whether the table rows are sorted according to the
   * domain attribute (before creating a {@link Dataset}).
   */
  @Override
  public void setSortDomainAxis(boolean sortDomainAxis){
    dummyTableChartStyle.setSortDomainAxis(sortDomainAxis);
  }

  /**
   * Returns whether the table rows are sorted according to the
   * domain attribute (before creating a {@link Dataset}).
   * @return {@code false} as default
   */
  @Override
  public boolean isSortDomainAxis() {
    return dummyTableChartStyle.isSortDomainAxis();
  }

  /** 
   * Sets whether a {@link CategoryDataset} is forced for
   * a numeric domain attribute. The default is to create {@link XYDataset}
   * for a numeric and {@link CategoryDataset} for a non-numeric domain
   * attribute.
   */
  @Override
  public void setForceCategories(boolean forceCategories){
    dummyTableChartStyle.setForceCategories(forceCategories);
  }

  /** 
   * Returns whether a {@link CategoryDataset} is forced for
   * a numeric domain attribute. The default is to create {@link XYDataset}
   * for a numeric and {@link CategoryDataset} for a non-numeric domain
   * attribute.
   * @return {@code false} as default
   */
  @Override
  public boolean isForceCategories(){
    return dummyTableChartStyle.isForceCategories();
  }

  /** 
   * Sets whether the attribute data is normalized for an
   * attribute(before creating a {@link Dataset}).
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @param normalize indicates the normalize property
   */  
  @Override
  public void setAttributeNormalized(int idx, Boolean normalize){
    dummyTableChartStyle.setAttributeNormalized(idx, normalize);
  }

  /** 
   * Returns whether the attribute data is normalized for an
   * attribute(before creating a {@link Dataset}).
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @return {@code false} as default
   */  
  @Override
  public boolean isAttributeNormalized(int idx) {
       return dummyTableChartStyle.isAttributeNormalized(idx);
  }

  /** 
   * Sets an arithmetical function which is calculated on
   * the attribute values (when creating a {@link Dataset}
   * e.g. from Features).
   * The function is calculated on all values with the same
   * domain value (X/category). If no function is set, duplicate
   * domain values are ignored (only the "last" value is shown, because
   * of replacement behavior).
   * @param idx attribute index (1=1st range series;
   *            2=2nd range series; ...)
   * @param func defines the calculated function
   * @exception IllegalArgumentException if function is set on
   *            domain attribute (0)
   */  
  @Override
  public void setAttributeAggregation(int idx, AggregationFunction func) {
    dummyTableChartStyle.setAttributeAggregation(idx, func);
  }
    
  /** 
   * Returns an arithmetical function which is calculated on
   * the attribute values (when creating a {@link Dataset}
   * e.g. from Features).
   * The function is calculated on all values with the same
   * domain value (X/category). If no function is set, duplicate
   * domain values are ignored (only the "last" value is shown, because
   * of replacement behavior).
   * @param idx attribute index (1=1st range series;
   *            2=2nd range series; ...)
   * @return {@code null} for index 0 (domain axis)
   */
  @Override
  public AggregationFunction getAttributeAggregation(int idx) {
   return dummyTableChartStyle.getAttributeAggregation(idx); 
  }

  /**
   * Sets the values, which are interpreted as "No Data". 
   * @param idx attribute index the "No Data" values are set for
   * @param noDataValues the "No Data" values
   */
  public void setNoDataValues(int idx, Set<Object> noDataValues) {
    dummyTableChartStyle.setNoDataValues(idx, noDataValues);
  }
  
  /**
   * Returns the values, which are interpreted as "No Data". 
   * @param idx attribute index the "No Data" values are returned for
   */
  public Set<Object> getNoDataValues(int idx) {
    return dummyTableChartStyle.getNoDataValues(idx);
  }
  
  /**
   * Sets a value, which is interpreted as "No Data". 
   * @param idx attribute index the "No Data" value is set for
   * @param noDataValue the "No Data" value
   */
  public void addNoDataValue(int idx, Object noDataValue) {
    dummyTableChartStyle.addNoDataValue(idx, noDataValue);
  }

  /**
   * Removes a "No Data" value for an attribute. 
   * @param idx attribute index the "No Data" value is removed for
   * @param noDataValue the "No Data" value to remove
   * @return {@code false} if the value was not an "No Data" value 
   */
  public boolean removeNoDataValue(int idx, Object noDataValue) {
    return dummyTableChartStyle.removeNoDataValue(idx, noDataValue);
  }

  /**
   * Checks whether the given value is one of the "No data" values for
   * the attribute.
   * @param idx attribute index the "No Data" value is checked for
   * @param value an attribute value
   */
  public boolean isNoDataValue(int idx, Object value) {
    return dummyTableChartStyle.isNoDataValue(idx, value);
  }

  /**
   * Checks whether the given value is one of the "No data" values for
   * the attribute. In this case this method returns {@code null}, otherwise
   * the value itself.
   * @param idx attribute index the "No Data" value is checked for
   * @param value an attribute value
   * @return {@code null} if the given value is one of the "No data" values 
   */
  public <T> T filterNoDataValue(int idx, T value) {
    return dummyTableChartStyle.filterNoDataValue(idx, value);
  }

  /**
   * Sets whether nodata values on the range axes are ignored when creating
   * the dataset(s). When {@code false}, these values are treat as numeric 0,
   * which can be result in poor peeks (e.g. in line charts).
   * For bar charts it can be useful to deal with NULLs as 0.<br>
   * Default: {@code true}   
   */
  @Override
  public void setIgnoreNoDataRangeValues(boolean ignoreNulls) {
    dummyTableChartStyle.setIgnoreNoDataRangeValues(ignoreNulls);
  }
  
  /**
   * Returns whether nodata values on the range axes are ignored when creating
   * the dataset(s). When {@code false}, these values are treat as numeric 0,
   * which can be result in poor peeks (e.g. in line charts).
   * For bar charts it can be useful to deal with NULLs as 0.<br>
   * Default: {@code true} 
   */
  @Override
  public boolean isIgnoreNoDataRangeValues() {
    return dummyTableChartStyle.isIgnoreNoDataRangeValues();
  }


  /**
   * Sets the table attribute the i-th chart attribute is weighted with.<br>
   * <b>Note:</b>
   * <ul>
   * <li>The weight attribute can not be set for the domain attribute 0.</li>
   * <li>Do not forget to set the NULL values for the weight attributes, too!</li>
   * </ul>
   * @param idx the table attribute
   * @param weightAttrName name of the weight attribute
   * @see #setWeightAttributeNoDataValues(idx, Set) to set NoDataValues for this
   *      attribute.<br/>
   * @exception IllegalArgumentException if weight attribute is set for
   *              attribute 0
   **/
  @Override
  public void setAttributeAggregationWeightAttributeName(int idx,
      String weightAttrName) {
    dummyTableChartStyle.setAttributeAggregationWeightAttributeName(idx, weightAttrName);
  }

  /**
   * Returns the table attribute the i-th chart attribute is weighted with.
   * <b>Note:</b><br>
   * This method returns {@code null} unless a
   * {@linkplain AggregationFunction#isWeighted() weighted} aggregation
   * function is set for the i-th attribute.
   * @param idx the table attribute
   * @see #setAttributeAggregationWeightAttributeName(int, String)
   **/
  @Override
  public String getAttributeAggregationWeightAttributeName(int idx) {
    return dummyTableChartStyle.getAttributeAggregationWeightAttributeName(idx);
  }

  /**
   * Sets the values, which are interpreted as "No Data" for the optional weight
   * attribute.
   * @param idx weight attribute index the "No Data" values are set for
   * @param noDataValues the "No Data" values
   */
  @Override
  public void setWeightAttributeNoDataValues(int idx, Set<Object> noDataValues) {
    dummyTableChartStyle.setWeightAttributeNoDataValues(idx, noDataValues);
  }

  /**
   * Returns the values, which are interpreted as "No Data" for the
   * weight attribute.
   * @param idx weight attribute index the "No Data" values are returned for
   */
  @Override
  public Set<Object> getWeightAttributeNoDataValues(int idx) {
    return dummyTableChartStyle.getWeightAttributeNoDataValues(idx);
  }

  /**
   * Sets a value, which is interpreted as "No Data" for the weight
   * attribute.
   * @param idx attribute index the "No Data" value is set for
   * @param noDataValue the "No Data" value
   */
  @Override
  public void addWeightAttributeNoDataValue(int idx, Object noDataValue) {
    dummyTableChartStyle.addWeightAttributeNoDataValue(idx, noDataValue);
  }

  /**
   * Removes a "No Data" value for a weight attribute.
   * @param idx attribute index the "No Data" value is removed for
   * @param noDataValue the "No Data" value to remove
   * @return {@code false} if the value was not an "No Data" value
   */
  @Override
  public boolean removeWeightAttributeNoDataValue(int idx, Object noDataValue) {
    return dummyTableChartStyle.removeWeightAttributeNoDataValue(idx, noDataValue);
  }

  /**
   * Checks whether the given value is one of the "No data" values for the
   * weight attribute.
   * @param idx weight attribute index the "No Data" value is checked for
   * @param value an attribute value
   * @return {@code false} is not weight attribute is set for attribute i
   * @see #setAttributeAggregationWeightAttributeName(int, String)
   */
  @Override
  public boolean isWeightAttributeNoDataValue(int idx, Object value) {
    return dummyTableChartStyle.isWeightAttributeNoDataValue(idx, value);
  }

  /**
   * Checks whether the given value is one of the "No data" values for the
   * attribute. In this case this method returns {@code null}, otherwise the
   * value itself.
   * @param idx attribute index the "No Data" value is checked for
   * @param value an attribute value
   * @return {@code null} if the given value is one of the "No data" values
   */
  @Override
  public <T> T filterWeightAttributeNoDataValue(int idx, T value) {
    return dummyTableChartStyle.filterWeightAttributeNoDataValue(idx, value);
  }

  /**
   * Defines the attribute whose values are used to define
   * groups. For each group value a series is created in the chart
   * dataset.
   * @param attrName table attribute name
   */
  @Override
  public void setSeriesAttributeName(String attrName) {
    dummyTableChartStyle.setSeriesAttributeName(attrName);
  }
  
  /**
   * Returns the attribute whose values are used to define
   * groups. For each group value a series is created in the chart
   * dataset.
   */
  @Override
  public String getSeriesAttributeName() {
    return dummyTableChartStyle.getSeriesAttributeName();
  }

  /**
   * Sets the values, which are interpreted as "No Data" for the
   * series attribute.
   * @param noDataValues the "No Data" values
   */
  @Override
  public void setSeriesAttributeNoDataValues(Set<Object> noDataValues) {
    dummyTableChartStyle.setSeriesAttributeNoDataValues(noDataValues);
  }

  /**
   * Returns the values, which are interpreted as "No Data" for the
   * series attribute.
   */
  @Override
  public Set<Object> getSeriesAttributeNoDataValues() {
    return dummyTableChartStyle.getSeriesAttributeNoDataValues();
  }

  /**
   * Sets a value, which is interpreted as "No Data" for the
   * series attribute.
   * @param noDataValue the "No Data" value
   */
  @Override
  public void addSeriesAttributeNoDataValue(Object noDataValue) {
    dummyTableChartStyle.addSeriesAttributeNoDataValue(noDataValue);
  }

  /**
   * Removes a "No Data" value for the series attribute.
   * @param noDataValue the "No Data" value to remove
   * @return {@code false} if the value was not an "No Data" value
   */
  @Override
 public boolean removeSeriesAttributeNoDataValue(Object noDataValue) {
    return dummyTableChartStyle.removeSeriesAttributeNoDataValue(noDataValue);
  }

  /**
   * Checks whether the given value is one of the "No data" values for the
   * series attribute.
   * @param value an attribute value
   */
  @Override
  public boolean isSeriesAttributeNoDataValue(Object value) {
    return dummyTableChartStyle.isSeriesAttributeNoDataValue(value);
  }

  /**
   * Checks whether the given value is one of the "No data" values for the
   * series attribute. In this case this method returns {@code null}, otherwise the
   * value itself.
   * @param value an attribute value
   * @return {@code null} if the given value is one of the "No data" values
   */
  @Override
  public <T> T filterSeriesAttributeNoDataValue(T value) {
    return dummyTableChartStyle.filterSeriesAttributeNoDataValue(value);
  }

  /**
   * Defines the attribute whose values are used to define the legend
   * title for the series defined by {@link #setSeriesAttributeName(String)}.
   * @param attrName table attribute name
   */
  public void setSeriesLegendTitleAttributeName(String attrName) {
    dummyTableChartStyle.setSeriesLegendTitleAttributeName(attrName);
  }
  
  /**
   * Returns the attribute whose values are used to define the legend
   * title for the series defined by {@link #setSeriesAttributeName(String)}.
   */
  public String getSeriesLegendTitleAttributeName() {
    return dummyTableChartStyle.getSeriesLegendTitleAttributeName();
  }
  
  /**
   * Creates a chart according to the given table data.
   * @param table the table
   * @exception UnsupportedOperationException if the style can not be applied to
   *            the table
   */
  @Override
  public abstract JFreeChart applyToTable(E table);
}
